from __future__ import annotations

import os
import shutil
from typing import Union, Generator, Optional, Tuple, Callable, Any, Set, TYPE_CHECKING
import warnings
import traceback
import numpy
import itertools
import inspect

from amulet import log
from amulet.api.block import Block, UniversalAirBlock
from amulet.api.block_entity import BlockEntity
from amulet.api.entity import Entity
from amulet.api.registry import BlockManager
from amulet.api.registry.biome_manager import BiomeManager
from amulet.api.errors import ChunkDoesNotExist, ChunkLoadError, LevelDoesNotExist
from amulet.api.chunk import Chunk
from amulet.api.selection import SelectionGroup, SelectionBox
from amulet.api.paths import get_temp_dir
from amulet.api.data_types import (
    OperationType,
    Dimension,
    VersionIdentifierType,
    BlockCoordinates,
    FloatTriplet,
    ChunkCoordinates,
)
from amulet.utils.generator import generator_unpacker
from amulet.utils.world_utils import block_coords_to_chunk_coords
from .chunk_manager import ChunkManager
from amulet.api.history.history_manager import MetaHistoryManager
from .clone import clone
import amulet.api.level

if TYPE_CHECKING:
    from PyMCTranslate import TranslationManager
    from amulet.api.wrapper import FormatWrapper


class BaseLevel:
    """
    BaseLevel handles chunk editing of any world or structure format via an separate and flexible data format.
    """

    def __init__(
        self, directory: str, format_wrapper: "FormatWrapper", temp_dir: str = None
    ):
        self._directory = directory
        if temp_dir is None:
            self._temp_directory = get_temp_dir(self._directory)
        else:
            self._temp_directory = temp_dir

        self._level_wrapper = format_wrapper
        self.level_wrapper.open()

        self._block_palette = BlockManager()
        self._block_palette.get_add_block(
            UniversalAirBlock
        )  # ensure that index 0 is always air

        self._biome_palette = BiomeManager()
        self._biome_palette.get_add_biome("universal_minecraft:plains")

        self._history_manager = MetaHistoryManager()

        self._chunks: ChunkManager = ChunkManager(
            os.path.join(self._temp_directory, "chunks"), self
        )

        self.history_manager.register(self._chunks, True)

    @property
    def level_wrapper(self) -> "FormatWrapper":
        """A class to access data directly from the level."""
        return self._level_wrapper

    @property
    def world_wrapper(self) -> "FormatWrapper":
        """A class to access data directly from the world."""
        warnings.warn(
            "BaseLevel.world_wrapper is depreciated and will be removed in the future. Please use BaseLevel.level_wrapper instead.",
            DeprecationWarning,
        )
        return self.level_wrapper

    @property
    def sub_chunk_size(self) -> int:
        """The normal dimensions of the chunk"""
        return self.level_wrapper.sub_chunk_size

    @property
    def level_path(self) -> str:
        """The system path where the level is located.
        This may be a directory or file depending on the level that is loaded."""
        return self._directory

    @property
    def world_path(self) -> str:
        """The directory where the world is located"""
        warnings.warn(
            "BaseLevel.world_path is depreciated and will be removed in the future. Please use BaseLevel.level_path instead.",
            DeprecationWarning,
        )
        return self._directory

    @property
    def translation_manager(self) -> "TranslationManager":
        """An instance of the translation class for use with this level."""
        return self.level_wrapper.translation_manager

    @property
    def palette(self) -> BlockManager:
        """The manager for the universal blocks in this level. New blocks must be registered here before adding to the level."""
        warnings.warn(
            "World.palette is depreciated and will be removed in the future. Please use BaseLevel.block_palette instead",
            DeprecationWarning,
        )
        return self.block_palette

    @property
    def block_palette(self) -> BlockManager:
        """The manager for the universal blocks in this level. New blocks must be registered here before adding to the level."""
        return self._block_palette

    @property
    def biome_palette(self) -> BiomeManager:
        """The manager for the universal blocks in this level. New biomes must be registered here before adding to the level."""
        return self._biome_palette

    @property
    def selection_bounds(self) -> SelectionGroup:
        """The selection(s) that all chunk data must fit within. Usually +/-30M for worlds. The selection for structures."""
        return self.level_wrapper.selection

    @property
    def dimensions(self) -> Tuple[Dimension, ...]:
        return tuple(self.level_wrapper.dimensions)

    def get_block(self, x: int, y: int, z: int, dimension: Dimension) -> Block:
        """
        Gets the universal Block object at the specified coordinates

        :param x: The X coordinate of the desired block
        :param y: The Y coordinate of the desired block
        :param z: The Z coordinate of the desired block
        :param dimension: The dimension of the desired block
        :return: The universal Block object representation of the block at that location
        :raises:
            ChunkDoesNotExist: If the chunk does not exist (was deleted or never created)
            ChunkLoadError: If the chunk was not able to be loaded. Eg. If the chunk is corrupt or some error occurred when loading.
        """
        cx, cz = block_coords_to_chunk_coords(x, z, sub_chunk_size=self.sub_chunk_size)
        offset_x, offset_z = x - 16 * cx, z - 16 * cz

        return self.get_chunk(cx, cz, dimension).get_block(offset_x, y, offset_z)

    def _chunk_box(
        self,
        cx: int,
        cz: int,
        sub_chunk_size: Optional[int] = None,
    ):
        """Get a SelectionBox containing the whole of a given chunk"""
        if sub_chunk_size is None:
            sub_chunk_size = self.sub_chunk_size
        return SelectionBox.create_chunk_box(cx, cz, sub_chunk_size)

    def get_coord_box(
        self,
        dimension: Dimension,
        selection: Union[SelectionGroup, SelectionBox, None] = None,
        yield_missing_chunks=False,
    ) -> Generator[Tuple[ChunkCoordinates, SelectionBox], None, None]:
        """Given a selection will yield chunk coordinates and `SelectionBox`es into that chunk
        If not given a selection will use the bounds of the object.

        :param selection: SelectionGroup or SelectionBox into the level
        :param dimension: The dimension to take effect in
        :param yield_missing_chunks: If a chunk does not exist an empty one will be created (defaults to false). Use this with care.
        """
        if isinstance(selection, SelectionBox):
            selection = SelectionGroup(selection)
        elif selection is None:
            selection = self.selection_bounds
        elif not isinstance(selection, SelectionGroup):
            raise TypeError(f"Expected a SelectionGroup but got {type(selection)}")

        selection: SelectionGroup
        if yield_missing_chunks or selection.footprint_area < 1_000_000:
            if yield_missing_chunks:
                for coord, box in selection.chunk_boxes(self.sub_chunk_size):
                    yield coord, box
            else:
                for (cx, cz), box in selection.chunk_boxes(self.sub_chunk_size):
                    if self.has_chunk(cx, cz, dimension):
                        yield (cx, cz), box

        else:
            # if the selection gets very large iterating over the whole selection and accessing chunks can get slow
            # instead we are going to iterate over the chunks and get the intersection of the selection
            for cx, cz in self.all_chunk_coords(dimension):
                box = SelectionGroup(
                    SelectionBox.create_chunk_box(cx, cz, self.sub_chunk_size)
                )

                if selection.intersects(box):
                    chunk_selection = selection.intersection(box)
                    for sub_box in chunk_selection.selection_boxes:
                        yield (cx, cz), sub_box

    def get_chunk_boxes(
        self,
        dimension: Dimension,
        selection: Union[SelectionGroup, SelectionBox, None] = None,
        create_missing_chunks=False,
    ) -> Generator[Tuple[Chunk, SelectionBox], None, None]:
        """Given a selection will yield chunks and `SelectionBox`es into that chunk
        If not given a selection will use the bounds of the object.

        :param selection: SelectionGroup or SelectionBox into the level
        :param dimension: The dimension to take effect in
        :param create_missing_chunks: If a chunk does not exist an empty one will be created (defaults to false). Use this with care.
        """
        for (cx, cz), box in self.get_coord_box(
            dimension, selection, create_missing_chunks
        ):
            try:
                chunk = self.get_chunk(cx, cz, dimension)
            except ChunkDoesNotExist:
                if create_missing_chunks:
                    yield self.create_chunk(cx, cz, dimension), box
            except ChunkLoadError:
                log.error(f"Error loading chunk\n{traceback.format_exc()}")
            else:
                yield chunk, box

    def get_chunk_slice_box(
        self,
        dimension: Dimension,
        selection: Union[SelectionGroup, SelectionBox],
        create_missing_chunks=False,
    ) -> Generator[Tuple[Chunk, Tuple[slice, slice, slice], SelectionBox], None, None]:
        """Given a selection will yield chunks, slices into that chunk and the corresponding box

        :param selection: SelectionGroup or SelectionBox into the level
        :param dimension: The dimension to take effect in
        :param create_missing_chunks: If a chunk does not exist an empty one will be created (defaults to false)
        Usage:
        for chunk, slice, box in level.get_chunk_slices(selection):
            chunk.blocks[slice] = ...
        """
        for chunk, box in self.get_chunk_boxes(
            dimension, selection, create_missing_chunks
        ):
            slices = box.chunk_slice(chunk.cx, chunk.cz, self.sub_chunk_size)
            yield chunk, slices, box

    def get_moved_coord_slice_box(
        self,
        dimension: Dimension,
        destination_origin: Tuple[int, int, int],
        selection: Optional[Union[SelectionGroup, SelectionBox]] = None,
        destination_sub_chunk_shape: Optional[int] = None,
        yield_missing_chunks: bool = False,
    ) -> Generator[
        Tuple[
            ChunkCoordinates,
            Tuple[slice, slice, slice],
            SelectionBox,
            ChunkCoordinates,
            Tuple[slice, slice, slice],
            SelectionBox,
        ],
        None,
        None,
    ]:
        """Iterate over a selection and return slices into the source object and destination object
        given the origin of the destination. When copying a selection to a new area the slices will
        only be equal if the offset is a multiple of the chunk size. This will rarely be the case
        so the slices need to be split up into parts that intersect a chunk in the source and destination.
        :param dimension: The dimension to iterate over.
        :param destination_origin: The location where the minimum point of self.selection_bounds will end up
        :param selection: An optional selection. The overlap of this and self.selection_bounds will be used
        :param destination_sub_chunk_shape: the chunk shape of the destination object (defaults to self.sub_chunk_size)
        :param yield_missing_chunks: Generate empty chunks if the chunk does not exist.
        :return:
        """
        if destination_sub_chunk_shape is None:
            destination_sub_chunk_shape = self.sub_chunk_size

        if selection is None:
            selection = self.selection_bounds
        else:
            selection = self.selection_bounds.intersection(selection)
        # the offset from self.selection to the destination location
        offset = numpy.subtract(
            destination_origin, self.selection_bounds.min, dtype=int
        )
        for (src_cx, src_cz), box in self.get_coord_box(
            dimension, selection, yield_missing_chunks=yield_missing_chunks
        ):
            dst_full_box = SelectionBox(
                offset + box.min,
                offset + box.max,
            )

            first_chunk = block_coords_to_chunk_coords(
                dst_full_box.min_x,
                dst_full_box.min_z,
                sub_chunk_size=destination_sub_chunk_shape,
            )
            last_chunk = block_coords_to_chunk_coords(
                dst_full_box.max_x - 1,
                dst_full_box.max_z - 1,
                sub_chunk_size=destination_sub_chunk_shape,
            )
            for dst_cx, dst_cz in itertools.product(
                range(first_chunk[0], last_chunk[0] + 1),
                range(first_chunk[1], last_chunk[1] + 1),
            ):
                chunk_box = self._chunk_box(dst_cx, dst_cz, destination_sub_chunk_shape)
                dst_box = chunk_box.intersection(dst_full_box)
                src_box = SelectionBox(-offset + dst_box.min, -offset + dst_box.max)
                src_slices = src_box.chunk_slice(src_cx, src_cz, self.sub_chunk_size)
                dst_slices = dst_box.chunk_slice(dst_cx, dst_cz, self.sub_chunk_size)
                yield (src_cx, src_cz), src_slices, src_box, (
                    dst_cx,
                    dst_cz,
                ), dst_slices, dst_box

    def get_moved_chunk_slice_box(
        self,
        dimension: Dimension,
        destination_origin: Tuple[int, int, int],
        selection: Optional[Union[SelectionGroup, SelectionBox]] = None,
        destination_sub_chunk_shape: Optional[int] = None,
        create_missing_chunks: bool = False,
    ) -> Generator[
        Tuple[
            Chunk,
            Tuple[slice, slice, slice],
            SelectionBox,
            ChunkCoordinates,
            Tuple[slice, slice, slice],
            SelectionBox,
        ],
        None,
        None,
    ]:
        """Iterate over a selection and return slices into the source object and destination object
        given the origin of the destination. When copying a selection to a new area the slices will
        only be equal if the offset is a multiple of the chunk size. This will rarely be the case
        so the slices need to be split up into parts that intersect a chunk in the source and destination.
        :param dimension: The dimension to iterate over.
        :param destination_origin: The location where the minimum point of self.selection will end up
        :param selection: An optional selection. The overlap of this and self.selection will be used
        :param destination_sub_chunk_shape: the chunk shape of the destination object (defaults to self.sub_chunk_size)
        :param create_missing_chunks: Generate empty chunks if the chunk does not exist.
        :return:
        """
        for (
            (src_cx, src_cz),
            src_slices,
            src_box,
            (dst_cx, dst_cz),
            dst_slices,
            dst_box,
        ) in self.get_moved_coord_slice_box(
            dimension,
            destination_origin,
            selection,
            destination_sub_chunk_shape,
            create_missing_chunks,
        ):
            try:
                chunk = self.get_chunk(src_cx, src_cz, dimension)
            except ChunkDoesNotExist:
                chunk = self.create_chunk(dst_cx, dst_cz, dimension)
            except ChunkLoadError:
                log.error(f"Error loading chunk\n{traceback.format_exc()}")
                continue
            yield chunk, src_slices, src_box, (dst_cx, dst_cz), dst_slices, dst_box

    def pre_save_operation(self) -> Generator[float, None, bool]:
        """Logic to run before saving. Eg recalculating height maps or lighting.
        Is a generator yielding progress from 0 to 1 and returning a bool saying if changes have been made.
        :return: Have any modifications been made.
        """
        return self.level_wrapper.pre_save_operation(self)

    def save(
        self,
        wrapper: "FormatWrapper" = None,
        progress_callback: Callable[[int, int], None] = None,
    ):
        """Save the level using the given wrapper.
        Leave as None to save back to the input wrapper.
        Optional progress callback to let the calling program know the progress. Input format chunk_index, chunk_count"""
        for chunk_index, chunk_count in self.save_iter(wrapper):
            if progress_callback is not None:
                progress_callback(chunk_index, chunk_count)

    def save_iter(
        self, wrapper: "FormatWrapper" = None
    ) -> Generator[Tuple[int, int], None, None]:
        """Save the level data back to the wrapper.

        :param wrapper: If specified will save the data to this wrapper instead of self.level_wrapper
        :return: A generator of the number of chunks completed and the total number of chunks
        """
        # TODO change the yield type to match OperationReturnType

        chunk_index = 0

        changed_chunks = list(self._chunks.changed_chunks())
        chunk_count = len(changed_chunks)

        if wrapper is None:
            wrapper = self.level_wrapper

        output_dimension_map = wrapper.dimensions

        # perhaps make this check if the directory is the same rather than if the class is the same
        save_as = wrapper is not self.level_wrapper
        if save_as:
            # The input wrapper is not the same as the loading wrapper (save-as)
            # iterate through every chunk in the input level and save them to the wrapper
            log.info(
                f"Converting level {self.level_wrapper.path} to level {wrapper.path}"
            )
            wrapper.translation_manager = (
                self.level_wrapper.translation_manager
            )  # TODO: this might cause issues in the future
            for dimension in self.level_wrapper.dimensions:
                chunk_count += len(list(self.level_wrapper.all_chunk_coords(dimension)))

            for dimension in self.level_wrapper.dimensions:
                try:
                    if dimension not in output_dimension_map:
                        continue
                    for cx, cz in self.level_wrapper.all_chunk_coords(dimension):
                        log.info(f"Converting chunk {dimension} {cx}, {cz}")
                        try:
                            chunk = self.level_wrapper.load_chunk(cx, cz, dimension)
                            wrapper.commit_chunk(chunk, dimension)
                        except ChunkLoadError:
                            log.info(f"Error loading chunk {cx} {cz}", exc_info=True)
                        chunk_index += 1
                        yield chunk_index, chunk_count
                        if not chunk_index % 10000:
                            wrapper.save()
                            self.level_wrapper.unload()
                            wrapper.unload()
                except LevelDoesNotExist:
                    continue

        for dimension, cx, cz in changed_chunks:
            if dimension not in output_dimension_map:
                continue
            try:
                chunk = self.get_chunk(cx, cz, dimension)
            except ChunkDoesNotExist:
                wrapper.delete_chunk(cx, cz, dimension)
            except ChunkLoadError:
                pass
            else:
                wrapper.commit_chunk(chunk, dimension)
                chunk.changed = False
            chunk_index += 1
            yield chunk_index, chunk_count
            if not chunk_index % 10000:
                wrapper.save()
                wrapper.unload()

        self.history_manager.mark_saved()
        log.info(f"Saving changes to level {wrapper.path}")
        wrapper.save()
        log.info(f"Finished saving changes to level {wrapper.path}")

    def purge(self):
        """Unload all loaded and cached data.
        This is functionally the same as closing and reopening the world without creating a new class."""
        self.unload()
        # self.history_manager.

    def close(self):
        """Close the attached level and remove temporary files
        Use changed method to check if there are any changes that should be saved before closing."""
        shutil.rmtree(self._temp_directory, ignore_errors=True)
        self.level_wrapper.close()

    def unload(self, safe_area: Optional[Tuple[Dimension, int, int, int, int]] = None):
        """Unload all chunks not in the safe area
        Safe area format: dimension, min chunk X|Z, max chunk X|Z"""
        self._chunks.unload(safe_area)
        self.level_wrapper.unload()

    def unload_unchanged(self):
        """Unload all data that has not been marked as changed."""
        self._chunks.unload_unchanged()

    @property
    def chunks(self) -> ChunkManager:
        """The object where chunk data is stored.
        There are methods in BaseLevel (this class) that can be used to modify chunk data."""
        return self._chunks

    def all_chunk_coords(self, dimension: Dimension) -> Set[Tuple[int, int]]:
        """The coordinates of every chunk in this dimension of the level.
        This is the combination of chunks saved to the level and chunks yet to be saved."""
        return self._chunks.all_chunk_coords(dimension)

    def has_chunk(self, cx: int, cz: int, dimension: Dimension) -> bool:
        """Does the chunk exist. This is a quick way to check if the chunk exists without loading it.

        :param cx: The x coordinate of the chunk.
        :param cz: The z coordinate of the chunk.
        :param dimension: The dimension to load the chunk from.
        :return: True if the chunk exists. Calling get_chunk on this chunk may still throw ChunkLoadError
        """
        return self._chunks.has_chunk(dimension, cx, cz)

    def get_chunk(self, cx: int, cz: int, dimension: Dimension) -> Chunk:
        """
        Gets the chunk data of the specified chunk coordinates.
        If the chunk does not exist ChunkDoesNotExist is raised.
        If some other error occurs then ChunkLoadError is raised (this error will also catch ChunkDoesNotExist)

        :param cx: The X coordinate of the desired chunk
        :param cz: The Z coordinate of the desired chunk
        :param dimension: The dimension to get the chunk from
        :return: A Chunk object containing the data for the chunk
        :raises:
            ChunkDoesNotExist: If the chunk does not exist (was deleted or never created)
            ChunkLoadError: If the chunk was not able to be loaded. Eg. If the chunk is corrupt or some error occurred when loading.
        """
        return self._chunks.get_chunk(dimension, cx, cz)

    def create_chunk(self, cx: int, cz: int, dimension: Dimension) -> Chunk:
        chunk = Chunk(cx, cz)
        self.put_chunk(chunk, dimension)
        return chunk

    def put_chunk(self, chunk: Chunk, dimension: Dimension):
        """Add a chunk to the universal level database"""
        self._chunks.put_chunk(chunk, dimension)

    def delete_chunk(self, cx: int, cz: int, dimension: Dimension):
        """Delete a chunk from the universal level database"""
        self._chunks.delete_chunk(dimension, cx, cz)

    def extract_structure(
        self, selection: SelectionGroup, dimension: Dimension
    ) -> amulet.api.level.ImmutableStructure:
        """Extract the area in the SelectionGroup from the level as a new structure"""
        return amulet.api.level.ImmutableStructure.from_level(
            self, selection, dimension
        )

    def extract_structure_iter(
        self, selection: SelectionGroup, dimension: Dimension
    ) -> Generator[float, None, amulet.api.level.ImmutableStructure]:
        """Extract the area in the SelectionGroup from the level as a new structure"""
        level = yield from amulet.api.level.ImmutableStructure.from_level_iter(
            self, selection, dimension
        )
        return level

    def paste(
        self,
        src_structure: "BaseLevel",
        src_dimension: Dimension,
        src_selection: SelectionGroup,
        dst_dimension: Dimension,
        location: BlockCoordinates,
        scale: FloatTriplet = (1.0, 1.0, 1.0),
        rotation: FloatTriplet = (0.0, 0.0, 0.0),
        include_blocks: bool = True,
        include_entities: bool = True,
        skip_blocks: Tuple[Block, ...] = (),
        copy_chunk_not_exist: bool = False,
    ):
        """Paste a structure into this structure at the given location.
        Note this command may change in the future.
        :param src_structure: The structure to paste into this structure.
        :param src_dimension: The dimension of the source structure to copy from.
        :param src_selection: The selection to copy from the source structure.
        :param dst_dimension: The dimension to paste the structure into.
        :param location: The location where the centre of the structure will be in the level
        :param scale: The scale in the x, y and z axis. These can be negative to mirror.
        :param rotation: The rotation in degrees around each of the axis.
        :param include_blocks: Include blocks when pasting the structure.
        :param include_entities: Include entities when pasting the structure.
        :param skip_blocks: If a block matches a block in this list it will not be copied.
        :param copy_chunk_not_exist: If a chunk does not exist in the source should it be copied over as air. Always False where level is a World.
        :return:
        """
        return generator_unpacker(
            self.paste_iter(
                src_structure,
                src_dimension,
                src_selection,
                dst_dimension,
                location,
                scale,
                rotation,
                include_blocks,
                include_entities,
                skip_blocks,
                copy_chunk_not_exist,
            )
        )

    def paste_iter(
        self,
        src_structure: "BaseLevel",
        src_dimension: Dimension,
        src_selection: SelectionGroup,
        dst_dimension: Dimension,
        location: BlockCoordinates,
        scale: FloatTriplet = (1.0, 1.0, 1.0),
        rotation: FloatTriplet = (0.0, 0.0, 0.0),
        include_blocks: bool = True,
        include_entities: bool = True,
        skip_blocks: Tuple[Block, ...] = (),
        copy_chunk_not_exist: bool = False,
    ) -> Generator[float, None, None]:
        """Paste a structure into this structure at the given location.
        Note this command may change in the future.
        :param src_structure: The structure to paste into this structure.
        :param src_dimension: The dimension of the source structure to copy from.
        :param src_selection: The selection to copy from the source structure.
        :param dst_dimension: The dimension to paste the structure into.
        :param location: The location where the centre of the structure will be in the level
        :param scale: The scale in the x, y and z axis. These can be negative to mirror.
        :param rotation: The rotation in degrees around each of the axis.
        :param include_blocks: Include blocks when pasting the structure.
        :param include_entities: Include entities when pasting the structure.
        :param skip_blocks: If a block matches a block in this list it will not be copied.
        :param copy_chunk_not_exist: If a chunk does not exist in the source should it be copied over as air. Always False where level is a World.
        :return: A generator of floats from 0 to 1 with the progress of the paste operation.
        """
        yield from clone(
            src_structure,
            src_dimension,
            src_selection,
            self,
            dst_dimension,
            self.selection_bounds,
            location,
            scale,
            rotation,
            include_blocks,
            include_entities,
            skip_blocks,
            copy_chunk_not_exist,
        )

    def get_version_block(
        self,
        x: int,
        y: int,
        z: int,
        dimension: Dimension,
        version: VersionIdentifierType,
    ) -> Union[Tuple[Block, BlockEntity], Tuple[Entity, None]]:
        """
        Get a block at the specified location and convert it to the format of the version specified
        Note the odd return format. In most cases this will return (Block, None) or (Block, BlockEntity)
        but in select cases like item frames may return (Entity, None)
        :param x: The X coordinate of the desired block
        :param y: The Y coordinate of the desired block
        :param z: The Z coordinate of the desired block
        :param dimension: The dimension of the desired block
        :param version: The version to get the block converted to.
        :return: The block at the given location converted to the `version` format. Note the odd return format.
        :raises:
            ChunkDoesNotExist: If the chunk does not exist (was deleted or never created)
            ChunkLoadError: If the chunk was not able to be loaded. Eg. If the chunk is corrupt or some error occurred when loading.
        """
        cx, cz = block_coords_to_chunk_coords(x, z, sub_chunk_size=self.sub_chunk_size)
        chunk = self.get_chunk(cx, cz, dimension)
        offset_x, offset_z = x - 16 * cx, z - 16 * cz

        output, extra_output, _ = self.translation_manager.get_version(
            *version
        ).block.from_universal(
            chunk.get_block(offset_x, y, offset_z), chunk.block_entities.get((x, y, z))
        )
        return output, extra_output

    def set_version_block(
        self,
        x: int,
        y: int,
        z: int,
        dimension: Dimension,
        version: VersionIdentifierType,
        block: Block,
        block_entity: BlockEntity = None,
    ):
        """
        Convert the block and block_entity from the given version format to the universal format and set at the location
        :param x: The X coordinate of the desired block
        :param y: The Y coordinate of the desired block
        :param z: The Z coordinate of the desired block
        :param dimension: The dimension of the desired block
        :param version: The version to get the block converted from.
        :param block:
        :param block_entity:
        :return: The block at the given location converted to the `version` format. Note the odd return format.
        :raises:
            ChunkLoadError: If the chunk was not able to be loaded. Eg. If the chunk is corrupt or some error occurred when loading.
        """
        cx, cz = block_coords_to_chunk_coords(x, z, sub_chunk_size=self.sub_chunk_size)
        try:
            chunk = self.get_chunk(cx, cz, dimension)
        except ChunkDoesNotExist:
            chunk = self.create_chunk(cx, cz, dimension)
        offset_x, offset_z = x - 16 * cx, z - 16 * cz

        (
            universal_block,
            universal_block_entity,
            _,
        ) = self.translation_manager.get_version(*version).block.to_universal(
            block, block_entity
        )
        chunk.set_block(offset_x, y, offset_z, universal_block),
        if isinstance(universal_block_entity, BlockEntity):
            chunk.block_entities[(x, y, z)] = universal_block_entity
        elif (x, y, z) in chunk.block_entities:
            del chunk.block_entities[(x, y, z)]

    # def get_entities_in_box(
    #     self, box: "SelectionGroup"
    # ) -> Generator[Tuple[Coordinates, List[object]], None, None]:
    #     # TODO: some of this logic can probably be moved the chunk class and have this method call that
    #     # TODO: update this to use the newer entity API
    #     out_of_place_entities = []
    #     entity_map: Dict[Tuple[int, int], List[List[object]]] = {}
    #     for chunk, subbox in self.get_chunk_boxes(box):
    #         entities = chunk.entities
    #         in_box = list(filter(lambda e: e.location in subbox, entities))
    #         not_in_box = filter(lambda e: e.location not in subbox, entities)
    #
    #         in_box_copy = deepcopy(in_box)
    #
    #         entity_map[chunk.coordinates] = [
    #             not_in_box,
    #             in_box,
    #         ]  # First index is the list of entities not in the box, the second is for ones that are
    #
    #         yield chunk.coordinates, in_box_copy
    #
    #         if (
    #             in_box != in_box_copy
    #         ):  # If an entity has been changed, update the dictionary entry
    #             entity_map[chunk.coordinates][1] = in_box_copy
    #         else:  # Delete the entry otherwise
    #             del entity_map[chunk.coordinates]
    #
    #     for chunk_coords, entity_list_list in entity_map.items():
    #         chunk = self.get_chunk(*chunk_coords)
    #         in_place_entities = list(
    #             filter(
    #                 lambda e: chunk_coords
    #                 == entity_position_to_chunk_coordinates(e.location),
    #                 entity_list_list[1],
    #             )
    #         )
    #         out_of_place = filter(
    #             lambda e: chunk_coords
    #             != entity_position_to_chunk_coordinates(e.location),
    #             entity_list_list[1],
    #         )
    #
    #         chunk.entities = in_place_entities + list(entity_list_list[0])
    #
    #         if out_of_place:
    #             out_of_place_entities.extend(out_of_place)
    #
    #     if out_of_place_entities:
    #         self.add_entities(out_of_place_entities)
    #
    # def add_entities(self, entities):
    #     proper_entity_chunks = map(
    #         lambda e: (entity_position_to_chunk_coordinates(e.location), e,), entities,
    #     )
    #     accumulated_entities: Dict[Tuple[int, int], List[object]] = {}
    #
    #     for chunk_coord, ent in proper_entity_chunks:
    #         if chunk_coord in accumulated_entities:
    #             accumulated_entities[chunk_coord].append(ent)
    #         else:
    #             accumulated_entities[chunk_coord] = [ent]
    #
    #     for chunk_coord, ents in accumulated_entities.items():
    #         chunk = self.get_chunk(*chunk_coord)
    #
    #         chunk.entities += ents
    #
    # def delete_entities(self, entities):
    #     chunk_entity_pairs = map(
    #         lambda e: (entity_position_to_chunk_coordinates(e.location), e,), entities,
    #     )
    #
    #     for chunk_coord, ent in chunk_entity_pairs:
    #         chunk = self.get_chunk(*chunk_coord)
    #         entities = chunk.entities
    #         entities.remove(ent)
    #         chunk.entities = entities

    def run_operation(
        self, operation: OperationType, dimension: Dimension, *args, create_undo=True
    ) -> Any:
        try:
            out = operation(self, dimension, *args)
            if inspect.isgenerator(out):
                out: Generator
                out = generator_unpacker(out)
        except Exception as e:
            self.restore_last_undo_point()
            raise e
        if create_undo:
            self.create_undo_point()
        return out

    @property
    def history_manager(self) -> MetaHistoryManager:
        """The class that manages undoing and redoing changes."""
        return self._history_manager

    def create_undo_point(self, world=True, non_world=True):
        """Create a restore point for all the data that has changed."""
        self.history_manager.create_undo_point(world, non_world)

    @property
    def changed(self) -> bool:
        """Has any data been modified but not saved to disk"""
        return self.history_manager.changed or self.level_wrapper.changed

    def undo(self):
        """Undoes the last set of changes to the level"""
        self.history_manager.undo()

    def redo(self):
        """Redoes the last set of changes to the level"""
        self.history_manager.redo()

    def restore_last_undo_point(self):
        """Restore the level to the state it was when self.create_undo_point was called.
        If an operation errors there may be modifications made that did not get tracked.
        This will revert those changes."""
        self.history_manager.restore_last_undo_point()



from __future__ import annotations

from .anvil_2203 import (
    Anvil2203Interface,
)


class Anvil2529Interface(Anvil2203Interface):
    def __init__(self):
        super().__init__()
        self.features["long_array_format"] = "1.16"

    @staticmethod
    def minor_is_valid(key: int):
        return 2529 <= key < 2681


export = Anvil2529Interface



from __future__ import annotations

from .leveldb_15 import (
    LevelDB15Interface,
)


class LevelDB16Interface(LevelDB15Interface):
    def __init__(self):
        super().__init__()

        self.features["chunk_version"] = 16


export = LevelDB16Interface



from typing import TYPE_CHECKING, Any, Tuple, Union, List
import numpy

import amulet_nbt
from amulet.api.wrapper import Interface
from .section import ConstructionSection
from amulet.api.chunk import Chunk
from amulet.api.block import Block
from amulet.api.selection import SelectionBox
from amulet.level.loader import Translators
from amulet.api.data_types import AnyNDArray

if TYPE_CHECKING:
    from amulet.api.wrapper import Translator
    from PyMCTranslate import TranslationManager


class ConstructionInterface(Interface):
    def is_valid(self, key: Tuple) -> bool:
        return key[0] == "construction" and key[1] == 0

    def decode(
        self, cx: int, cz: int, data: List[ConstructionSection]
    ) -> Tuple["Chunk", AnyNDArray]:
        chunk = Chunk(cx, cz)
        palette = []
        for section in data:
            if any(s == 0 for s in section.shape):
                continue
            if section.blocks is not None:
                shapex, shapey, shapez = section.shape
                sx = section.sx % 16
                sy = section.sy
                sz = section.sz % 16
                chunk.blocks[
                    sx : sx + shapex,
                    sy : sy + shapey,
                    sz : sz + shapez,
                ] = section.blocks.astype(numpy.uint32) + len(palette)
                chunk.block_entities.update(section.block_entities)
            chunk.entities.extend(section.entities)
            palette += section.palette

        np_palette, inverse = numpy.unique(palette, return_inverse=True)
        np_palette = numpy.insert(
            np_palette,
            0,
            Block(
                namespace="minecraft",
                base_name="air",
                properties={"block_data": amulet_nbt.TAG_Int(0)},
            ),
        )
        inverse += 1
        np_palette: AnyNDArray
        inverse: numpy.ndarray
        for cy in chunk.blocks.sub_chunks:
            chunk.blocks.add_sub_chunk(
                cy, inverse[chunk.blocks.get_sub_chunk(cy)].astype(numpy.uint32)
            )
        return chunk, np_palette

    def encode(
        self,
        chunk: "Chunk",
        palette: AnyNDArray,
        max_world_version: Tuple[str, Union[int, Tuple[int, int, int]]],
        boxes: List[SelectionBox] = None,
    ) -> List[ConstructionSection]:
        sections = []
        boxes = boxes or []
        for box in boxes:
            cx, cz = chunk.cx, chunk.cz
            for cy in box.chunk_y_locations():
                sub_box = box.intersection(
                    SelectionBox.create_sub_chunk_box(cx, cy, cz)
                )
                entities = [e for e in chunk.entities if e.location in sub_box]
                if cy in chunk.blocks:
                    sections.append(
                        ConstructionSection(
                            sub_box.min,
                            sub_box.shape,
                            numpy.asarray(chunk.blocks[sub_box.chunk_slice(cx, cz)]),
                            list(palette),
                            entities,
                            [e for e in chunk.block_entities if e.location in sub_box],
                        )
                    )
                elif entities:
                    sections.append(
                        ConstructionSection(
                            sub_box.min, sub_box.shape, None, [], entities, []
                        )
                    )

        return sections

    def get_translator(
        self,
        max_world_version: Tuple[str, Tuple[int, int, int]],
        data: Any = None,
        translation_manager: "TranslationManager" = None,
    ) -> Tuple["Translator", Union[int, Tuple[int, int, int]]]:
        platform, version_number = max_world_version
        version = translation_manager.get_version(platform, version_number)
        if platform == "java":
            version_number = version.data_version
        return Translators.get((platform, version_number)), version_number


class Construction0Interface(ConstructionInterface):
    pass



from typing import List, Tuple

from amulet.api.level import BaseLevel
from amulet.api.data_types import Dimension


class StructureCache:
    """A class for storing and accessing structure objects"""

    def __init__(self):
        self._structure_buffer: List[Tuple[BaseLevel, Dimension]] = []

    def add_structure(self, structure: BaseLevel, dimension: Dimension):
        """Add a structure to the cache"""
        assert isinstance(
            structure, BaseLevel
        ), "structure given is not a BaseLevel instance"
        self._structure_buffer.append((structure, dimension))

    def get_structure(self, index=-1) -> Tuple[BaseLevel, Dimension]:
        """Get a structure from the cache. Default last."""
        return self._structure_buffer[index]

    def pop_structure(self, index=-1) -> Tuple[BaseLevel, Dimension]:
        """Get and remove a structure from the cache. Default last."""
        return self._structure_buffer.pop(index)

    def __len__(self) -> int:
        """Get the number of structures in the cache"""
        return len(self._structure_buffer)

    def clear(self):
        """Empty the cache."""
        self._structure_buffer.clear()


structure_cache = StructureCache()



from typing import Any, Generator, List, Dict, Tuple, Optional, TYPE_CHECKING

from amulet.api.data_types import Dimension, PlatformType, ChunkCoordinates
from amulet.api.wrapper import FormatWrapper
from amulet.api.errors import ChunkDoesNotExist

if TYPE_CHECKING:
    from amulet.api.wrapper import Interface


class VoidFormatWrapper(FormatWrapper):
    """There is no actual database here for chunks to be read from or written to. This is just here to make the world happy."""

    @staticmethod
    def is_valid(path: str) -> bool:
        return False

    @property
    def valid_formats(self) -> Dict[PlatformType, Tuple[bool, bool]]:
        return {}

    @property
    def dimensions(self) -> List[Dimension]:
        return ["main"]

    @property
    def can_add_dimension(self) -> bool:
        return False

    def register_dimension(self, dimension_internal: Any, dimension_name: Dimension):
        pass

    def _get_interface(self, raw_chunk_data: Optional[Any] = None) -> "Interface":
        raise Exception("If this is called something is wrong")

    def _create(self, overwrite: bool, **kwargs):
        pass

    def _open(self):
        pass

    def _save(self):
        pass

    def _close(self):
        pass

    def unload(self):
        pass

    def all_chunk_coords(
        self, dimension: Dimension
    ) -> Generator[ChunkCoordinates, None, None]:
        yield from ()

    def has_chunk(self, cx: int, cz: int, dimension: Dimension) -> bool:
        return False

    def _delete_chunk(self, cx: int, cz: int, dimension: Dimension):
        pass

    def _put_raw_chunk_data(self, cx: int, cz: int, data: Any, dimension: Dimension):
        pass

    def _get_raw_chunk_data(self, cx: int, cz: int, dimension: Dimension) -> Any:
        raise ChunkDoesNotExist



from __future__ import annotations

import itertools
import numpy

from typing import Tuple, Iterable, List, Generator, Union, Optional, overload

from amulet.api.data_types import (
    CoordinatesAny,
    ChunkCoordinates,
    SubChunkCoordinates,
    FloatTriplet,
    PointCoordinatesAny,
)
import amulet
from .box import SelectionBox


class SelectionGroup:
    """A container for zero or more `SelectionBox`es.
    This allows for non-rectangular and non-contiguous selections.
    """

    def __init__(
        self, selection_boxes: Union[SelectionBox, Iterable[SelectionBox]] = ()
    ):
        if type(selection_boxes) is SelectionBox:
            self._selection_boxes: Tuple[SelectionBox, ...] = (selection_boxes,)
        elif type(selection_boxes) in (tuple, list):
            self._selection_boxes: Tuple[SelectionBox, ...] = tuple(
                box for box in selection_boxes if type(box) is SelectionBox
            )
        else:
            amulet.log.warning(f"Invalid format for selection_boxes {selection_boxes}")
            self._selection_boxes: Tuple[SelectionBox, ...] = ()

    def __repr__(self):
        boxes = ", ".join([repr(box) for box in self.selection_boxes])
        return f"SelectionGroup([{boxes}])"

    def __str__(self):
        boxes = ", ".join([str(box) for box in self.selection_boxes])
        return f"[{boxes}]"

    def __eq__(self, other: SelectionGroup) -> bool:
        return (
            type(other) is SelectionGroup
            and self.selection_boxes_sorted == other.selection_boxes_sorted
        )

    def __add__(self, other: SelectionGroup) -> SelectionGroup:
        if not type(other) is SelectionGroup:
            return NotImplemented
        return SelectionGroup(self.selection_boxes + other.selection_boxes)

    def __iter__(self) -> Iterable[SelectionBox]:
        """A generator of all the boxes in the group."""
        yield from self._selection_boxes

    def __len__(self) -> int:
        """The number of selection boxes in the group."""
        return len(self._selection_boxes)

    def __contains__(self, item: CoordinatesAny) -> bool:
        """Is the block (int) or point (float) location within any of the boxes in this group."""
        return self.contains_block(item)

    def contains_block(self, coords: CoordinatesAny) -> bool:
        """Is the coordinate greater than or equal to the min point but less than the max point of any of the boxes."""
        return any(box.contains_block(coords) for box in self._selection_boxes)

    def contains_point(self, coords: CoordinatesAny) -> bool:
        """Is the coordinate greater than or equal to the min point but less than or equal to the max point of any of the boxes."""
        return any(box.contains_point(coords) for box in self._selection_boxes)

    def blocks(self) -> Iterable[Tuple[int, int, int]]:
        """An iterable of every block in the selection.
        Note: if boxes intersect, the blocks in the intersected region will be included multiple times.
        If this behaviour is not desired the `merge_boxes` method will return a new SelectionGroup with no intersections."""
        return itertools.chain.from_iterable(self.selection_boxes)

    def __bool__(self) -> bool:
        """Are there any boxes in the group."""
        return bool(self._selection_boxes)

    @overload
    def __getitem__(self, item: int) -> SelectionBox:
        ...

    @overload
    def __getitem__(self, item: slice) -> SelectionGroup:
        ...

    def __getitem__(self, item):
        """Get the selection box at the given index."""
        val = self._selection_boxes[item]
        if type(val) is tuple:
            return SelectionGroup(val)
        else:
            return val

    @property
    def min(self) -> numpy.ndarray:
        """The minimum point of of all the boxes in the group."""
        if self._selection_boxes:
            return numpy.min(numpy.array([box.min for box in self._selection_boxes]), 0)
        else:
            raise ValueError("SelectionGroup does not contain any SelectionBoxes")

    @property
    def max(self) -> numpy.ndarray:
        """The maximum point of of all the boxes in the group."""
        if self._selection_boxes:
            return numpy.max(numpy.array([box.max for box in self._selection_boxes]), 0)
        else:
            raise ValueError("SelectionGroup does not contain any SelectionBoxes")

    def to_box(self) -> SelectionBox:
        """Create a `SelectionBox` based off the bounds of the boxes in the group."""
        return SelectionBox(self.min, self.max)

    def merge_boxes(self) -> SelectionGroup:
        """Take the boxes as they were given to this class, merge neighbouring boxes and remove overlapping regions.
        The result should be a SelectionGroup containing one or more SelectionBox classes that represents the same
        volume as the original but with no overlapping boxes."""
        selection_boxes = self.selection_boxes
        # TODO remove duplicate boxes

        # remove duplicates
        selection_boxes_ = []
        for box in selection_boxes:
            if not any(box == box_ for box_ in selection_boxes_):
                selection_boxes_.append(box)
        selection_boxes = selection_boxes_

        if len(selection_boxes) >= 2:
            merge_boxes = True
            while merge_boxes:
                # find two neighbouring boxes and merge them
                merge_boxes = False  # if two boxes get merged this will be set back to True and this will run again.
                box_index = 0  # the index of the first box
                while box_index < len(selection_boxes):
                    box = selection_boxes[box_index]
                    other_index = box_index + 1  # the index of the second box.
                    # This always starts at one greater than box_index because
                    # the lower values were already checked the other way around
                    while other_index < len(selection_boxes):
                        other = selection_boxes[other_index]
                        x_dim = box.min_x == other.min_x and box.max_x == other.max_x
                        y_dim = box.min_y == other.min_y and box.max_y == other.max_y
                        z_dim = box.min_z == other.min_z and box.max_z == other.max_z

                        x_border = box.max_x == other.min_x or other.max_x == box.min_x
                        y_border = box.max_y == other.min_y or other.max_y == box.min_y
                        z_border = box.max_z == other.min_z or other.max_z == box.min_z

                        if (
                            (x_dim and y_dim and z_border)
                            or (x_dim and z_dim and y_border)
                            or (y_dim and z_dim and x_border)
                        ):
                            selection_boxes.pop(other_index)
                            selection_boxes.pop(box_index)
                            selection_boxes.append(
                                SelectionBox(
                                    numpy.min([box.min, other.min], 0),
                                    numpy.max([box.max, other.max], 0),
                                )
                            )
                            merge_boxes = True
                            box = selection_boxes[box_index]
                            other_index = box_index + 1
                        else:
                            other_index += 1
                    box_index += 1
        return SelectionGroup(selection_boxes)

    @property
    def is_contiguous(self) -> bool:
        """Does the SelectionGroup represent one connected region (True) or multiple separated regions (False).
        If two boxes are touching at the corners this is classed as contiguous."""
        # TODO: This needs some work. It will only work if the selections touch in a chain
        #  it does not care if it loops back and intersects itself. Does intersecting count as being contiguous?
        #  I would say yes
        if len(self._selection_boxes) == 1:
            return True

        for i in range(len(self._selection_boxes) - 1):
            sub_box: SelectionBox = self._selection_boxes[i]
            next_box: SelectionBox = self._selection_boxes[i + 1]
            if not sub_box.touches(next_box):
                return False

        return True

    @property
    def is_rectangular(self) -> bool:
        """
        Checks if the SelectionGroup is a rectangle

        :return: True is the selection is a rectangle, False otherwise
        """
        return (
            len(self._selection_boxes) == 1
            or len(self.merge_boxes().selection_boxes) == 1
        )

    @property
    def selection_boxes(self) -> List[SelectionBox]:
        """
        Returns a read only list of unmodified `SelectionBox`es in the SelectionGroup.
        :return: A list of the `SelectionBox`es
        """
        return list(self._selection_boxes)

    @property
    def selection_boxes_sorted(self) -> List[SelectionBox]:
        """
        Returns a list of unmodified `SelectionBox`es in the SelectionGroup sorted based on the hash of the coordinates.
        :return: A list of the `SelectionBox`es
        """
        return sorted(self._selection_boxes, key=hash)

    def chunk_locations(
        self, sub_chunk_size: int = 16
    ) -> Generator[ChunkCoordinates, None, None]:
        """A generator of chunk locations that the boxes in this group intersect.
        Each location is only given once even if there are multiple boxes in the chunk.
        :param sub_chunk_size: The dimension of the chunk (normally 16)
        """
        yield from set(
            location
            for box in self.selection_boxes
            for location in box.chunk_locations(sub_chunk_size)
        )

    def chunk_boxes(
        self, sub_chunk_size: int = 16
    ) -> Generator[Tuple[ChunkCoordinates, SelectionBox], None, None]:
        """A generator of modified `SelectionBox`es to fit within each chunk.
        If a box straddles multiple chunks this method will split it up into a box
        for each chunk it intersects along with the chunk coordinates of that chunk.
        :param sub_chunk_size: The dimension of the chunk (normally 16)
        """
        for box in self.selection_boxes:
            yield from box.chunk_boxes(sub_chunk_size)

    def sub_chunk_locations(
        self, sub_chunk_size: int = 16
    ) -> Generator[SubChunkCoordinates, None, None]:
        """A generator of sub-chunk locations that the boxes in this group intersect.
        Each location is only given once even if there are multiple boxes in the chunk.
        :param sub_chunk_size: The dimension of the chunk (normally 16)
        """
        yield from set(
            location
            for box in self.selection_boxes
            for location in box.sub_chunk_locations(sub_chunk_size)
        )

    def sub_chunk_boxes(
        self, sub_chunk_size: int = 16
    ) -> Generator[Tuple[SubChunkCoordinates, SelectionBox], None, None]:
        """A generator of modified `SelectionBox`es to fit within each sub-chunk.
        If a box straddles multiple sub-chunks this method will split it up into a box
        for each sub-chunk it intersects along with the sub-chunk coordinates of that sub-chunk.
        :param sub_chunk_size: The dimension of the chunk (normally 16)
        """
        for box in self.selection_boxes:
            for (cx, cy, cz), sub_box in box.sub_chunk_boxes(sub_chunk_size):
                yield (cx, cy, cz), box

    def intersects(self, other: Union[SelectionGroup, SelectionBox]) -> bool:
        """Check if self and other intersect."""
        if isinstance(other, SelectionGroup):
            return any(
                self_box.intersects(other_box)
                for self_box in self.selection_boxes
                for other_box in other.selection_boxes
            )
        elif isinstance(other, SelectionBox):
            return any(self_box.intersects(other) for self_box in self.selection_boxes)

    @staticmethod
    def _to_group(other: Union[SelectionGroup, SelectionBox]) -> SelectionGroup:
        if type(other) not in (SelectionGroup, SelectionBox):
            raise TypeError("other must be a SelectionGroup or SelectionBox.")
        if isinstance(other, SelectionBox):
            other = SelectionGroup(other)
        return other

    def intersection(
        self, other: Union[SelectionGroup, SelectionBox]
    ) -> SelectionGroup:
        """Get a new SelectionGroup that represents the area contained within self and other."""
        other = self._to_group(other)
        intersection = []
        for self_box in self.selection_boxes:
            for other_box in other.selection_boxes:
                if self_box.intersects(other_box):
                    intersection.append(self_box.intersection(other_box))
        return SelectionGroup(intersection)

    def subtract(self, other: Union[SelectionGroup, SelectionBox]) -> SelectionGroup:
        """Returns a new SelectionGroup containing the volume that does not intersect with other."""
        other = self._to_group(other)
        selections = self.selection_boxes
        for other_box in other.selection_boxes:
            # for each box in other
            selections_new = []
            for self_box in selections:
                selections_new += self_box.subtract(other_box)
            selections = selections_new
            if not selections:
                break
        return SelectionGroup(selections)

    def union(self, other: Union[SelectionGroup, SelectionBox]) -> SelectionGroup:
        """Returns a new SelectionGroup containing the volume of self and other."""
        other = self._to_group(other)
        if other.is_subset(self):
            return self
        else:
            return self.subtract(other) + other

    def is_subset(self, other: Union[SelectionGroup, SelectionBox]) -> bool:
        """Is self completely within other."""
        other = self._to_group(other)
        return not self.subtract(other)

    def closest_vector_intersection(
        self, origin: PointCoordinatesAny, vector: PointCoordinatesAny
    ) -> Tuple[Optional[int], float]:
        """
        Returns the index for the closest box in the look vector and the multiplier of the look vector to get there.
        :param origin: The origin tuple of the vector
        :param vector: The vector magnitude in x, y and z
        :return: Index for the closest box and the multiplier of the vector to get there. None, inf if no intersection.
        """
        index_return = None
        multiplier = float("inf")
        for index, box in enumerate(self._selection_boxes):
            mult = box.intersects_vector(origin, vector)
            if mult is not None and mult < multiplier:
                multiplier = mult
                index_return = index
        return index_return, multiplier

    def transform(
        self, scale: FloatTriplet, rotation: FloatTriplet, translation: FloatTriplet
    ) -> SelectionGroup:
        """creates a new transformed SelectionGroup."""
        selection_group = []
        for selection in self.selection_boxes:
            for transformed_selection in selection.transform(
                scale, rotation, translation
            ):
                selection_group.append(transformed_selection)
        return SelectionGroup(selection_group)

    def copy(self):
        return SelectionGroup(self.selection_boxes)

    @property
    def volume(self) -> int:
        """The volume of all the selection boxes combined."""
        return sum(box.volume for box in self.selection_boxes)

    @property
    def footprint_area(self) -> int:
        """The flat area of the selection."""
        return SelectionGroup(
            [
                SelectionBox((box.min_x, 0, box.min_z), (box.max_x, 1, box.max_z))
                for box in self.selection_boxes
            ]
        ).volume


if __name__ == "__main__":
    b1 = SelectionBox((0, 0, 0), (4, 4, 4))
    b2 = SelectionBox((7, 7, 7), (10, 10, 10))
    sel = SelectionGroup((b1, b2))

    for x, y, z in sel.blocks():
        print(x, y, z)



from __future__ import annotations

from .leveldb_14 import (
    LevelDB14Interface,
)


class LevelDB15Interface(LevelDB14Interface):
    def __init__(self):
        super().__init__()

        self.features["chunk_version"] = 15


export = LevelDB15Interface



from __future__ import annotations

from .leveldb_12 import (
    LevelDB12Interface,
)


class LevelDB13Interface(LevelDB12Interface):
    def __init__(self):
        super().__init__()

        self.features["chunk_version"] = 13


export = LevelDB13Interface



from __future__ import annotations

import math
import sys
import gzip
from io import StringIO
from typing import Tuple, Optional
import numpy
from numpy import ndarray, zeros, uint8
from amulet.api.data_types import ChunkCoordinates

SECTOR_BYTES = 4096
SECTOR_INTS = SECTOR_BYTES / 4
CHUNK_HEADER_SIZE = 5
VERSION_GZIP = 1
VERSION_DEFLATE = 2


def block_coords_to_chunk_coords(
    *args: int, sub_chunk_size: int = 16
) -> Tuple[int, ...]:
    """
    Converts the supplied block coordinates into chunk coordinates

    :param args: The coordinate of the block(s)
    :param sub_chunk_size: The dimension of the chunk (Optional. Default 16)
    :return: The resulting chunk coordinates in (x, z) order
    """
    return tuple(int(math.floor(coord / sub_chunk_size)) for coord in args)


def chunk_coords_to_block_coords(
    x: int, z: int, chunk_x_size: int = 16, chunk_z_size: int = 16
) -> ChunkCoordinates:
    """
    Converts the supplied chunk coordinates into block coordinates

    :param x: The x coordinate of the chunk
    :param z: The z coordinate of the chunk
    :param chunk_x_size: The dimension of the chunk in the x direction (Optional. Default 16)
    :param chunk_z_size: The dimension of the chunk in the z direction (Optional. Default 16)
    :return: The resulting block coordinates in (x, z) order
    """
    return x * chunk_x_size, z * chunk_z_size


def chunk_coords_to_region_coords(cx: int, cz: int) -> ChunkCoordinates:
    """
    Converts the supplied chunk coordinates into region coordinates

    :param cx: The x coordinate of the chunk
    :param cz: The z coordinate of the chunk
    :return: The resulting region coordinates in (x, z) order
    """
    return cx >> 5, cz >> 5


def region_coords_to_chunk_coords(rx: int, rz: int) -> ChunkCoordinates:
    """
    Converts the supplied region coordinates into the minimum chunk coordinates of that region

    :param rx: The x coordinate of the region
    :param rz: The y coordinate of the region
    :return: The resulting minimum chunk coordinates of that region in (x, z) order
    """
    return rx << 5, rz << 5


def blocks_slice_to_chunk_slice(
    blocks_slice: slice, chunk_shape: int, chunk_coord: int
) -> slice:
    """
    Converts the supplied blocks slice into chunk slice
    :param blocks_slice: The slice of the blocks
    :param chunk_shape: The shape of the chunk in this direction
    :param chunk_coord: The coordinate of the chunk in this direction
    :return: The resulting chunk slice
    """
    return slice(
        min(max(0, blocks_slice.start - chunk_coord * chunk_shape), chunk_shape),
        min(max(0, blocks_slice.stop - chunk_coord * chunk_shape), chunk_shape),
    )


def gunzip(data):
    """
    Decompresses data that is in Gzip format
    """
    return gzip.GzipFile(fileobj=StringIO(data)).read()


def from_nibble_array(arr: ndarray) -> ndarray:
    """
    Unpacks a flat nibble array into a full size numpy array

    :param arr: The nibble array
    :return: The resulting array
    """
    shape = arr.size

    new_arr = zeros((shape * 2), dtype=uint8)

    new_arr[::2] = arr & 0xF
    new_arr[1::2] = arr >> 4

    return new_arr


def to_nibble_array(arr: ndarray) -> ndarray:
    """
    packs a full size numpy array into a nibble array.

    :param arr: Full numpy array
    :return: The nibble array
    """
    arr = arr.ravel()
    return (arr[::2] + (arr[1::2] << 4)).astype("uint8")


def decode_long_array(
    long_array: numpy.ndarray, size: int, dense=True
) -> numpy.ndarray:
    """
    Decode an long array (from BlockStates or Heightmaps)
    :param long_array: Encoded long array
    :param size: int: The expected size of the returned array
    :return: Decoded array as numpy array
    """
    long_array = long_array.astype(">q")
    bits_per_entry = (len(long_array) * 64) // size
    bits = numpy.unpackbits(long_array[::-1].astype(">i8").view("uint8"))
    if not dense:
        entry_per_long = 64 // bits_per_entry
        bits = bits.reshape(-1, 64)[:, -entry_per_long * bits_per_entry :]

    return numpy.packbits(
        numpy.pad(
            bits.reshape(-1, bits_per_entry)[-size:, :],
            [(0, 0), (16 - bits_per_entry, 0)],
            "constant",
        )
    ).view(dtype=">h")[::-1]


def encode_long_array(
    array: numpy.ndarray,
    dense: bool = True,
    bits_per_entry: Optional[int] = None,
    min_bits_per_entry=4,
) -> numpy.ndarray:
    """
    Encode an long array (from BlockStates or Heightmaps)
    :param array: A numpy array of the data to be encoded.
    :param dense: If true the long arrays will be treated as a bit stream. If false they are distinct values with padding
    :param bits_per_entry: The number of bits to use to store each value. If left as None will use the smallest bits per entry.
    :return: Encoded array as numpy array
    """
    array = array.astype(">h")
    required_bits_per_entry = max(
        int(numpy.amax(array)).bit_length(), min_bits_per_entry
    )
    if bits_per_entry is None:
        bits_per_entry = required_bits_per_entry
    else:
        if required_bits_per_entry > bits_per_entry:
            raise Exception(
                f"The array requires at least {required_bits_per_entry} bits per value which is more than the specified {bits_per_entry} bits"
            )
    if not dense:
        if bits_per_entry == 11:
            bits_per_entry = 12  # 11 and 12 take up the same amount of space. I don't know if 11 exists any more.
    bits = numpy.unpackbits(numpy.ascontiguousarray(array[::-1]).view("uint8")).reshape(
        -1, 16
    )[:, -bits_per_entry:]
    if not dense:
        entry_per_long = 64 // bits_per_entry
        if bits.shape[0] % entry_per_long:
            bits = numpy.pad(
                bits,
                [(entry_per_long - (bits.shape[0] % entry_per_long), 0), (0, 0)],
                "constant",
            )
        bits = numpy.pad(
            bits.reshape(-1, bits_per_entry * entry_per_long),
            [(0, 0), (64 - bits_per_entry * entry_per_long, 0)],
            "constant",
        )

    return numpy.packbits(bits).view(dtype=">q")[::-1]


def get_size(obj, seen=None):
    """Recursively finds size of objects"""
    size = sys.getsizeof(obj)
    if seen is None:
        seen = set()
    obj_id = id(obj)
    if obj_id in seen:
        return 0

    # Important mark as seen *before* entering recursion to gracefully handle
    # self-referential objects
    seen.add(obj_id)
    if isinstance(obj, dict):
        size += sum([get_size(v, seen) for v in obj.values()])
        size += sum([get_size(k, seen) for k in obj.keys()])
    elif hasattr(obj, "__dict__"):
        size += get_size(obj.__dict__, seen)
    elif hasattr(obj, "__iter__") and not isinstance(obj, (str, bytes, bytearray)):
        size += sum([get_size(i, seen) for i in obj])
    return size


def get_smallest_dtype(arr: ndarray, uint: bool = True) -> int:
    """
    Returns the smallest dtype (number) that the array can afford
    :param arr: The array to check on
    :param uint: Should the array fit in uint or not (default: True)
    :return: The number of bits all the elements can be represented with
    """
    possible_dtypes = (2 ** x for x in range(3, 8))
    max_number = numpy.amax(arr)
    if not uint:
        max_number = max_number * 2
    if max_number == 0:
        max_number = 1
    return next(dtype for dtype in possible_dtypes if dtype > math.log(max_number, 2))


def entity_position_to_chunk_coordinates(
    entity_coordinates: Tuple[float, float, float]
):
    return (
        int(math.floor(entity_coordinates[0])) >> 4,
        int(math.floor(entity_coordinates[2])) >> 4,
    )


def fast_unique(array: numpy.ndarray) -> Tuple[numpy.ndarray, numpy.ndarray]:
    uni = numpy.unique(array)
    lut = numpy.zeros(numpy.amax(uni) + 1, dtype=numpy.uint32)
    lut[uni] = numpy.arange(uni.size)
    inv = lut[array]
    return uni, inv



from __future__ import annotations

from .leveldb_9 import (
    LevelDB9Interface,
)


class LevelDB10Interface(LevelDB9Interface):
    def __init__(self):
        super().__init__()

        self.features["chunk_version"] = 10


export = LevelDB10Interface



from __future__ import annotations

from .anvil_world import AnvilFormat
from amulet.utils.format_utils import check_all_exist, load_leveldat


class AnvilForgeFormat(AnvilFormat):
    @staticmethod
    def is_valid(directory) -> bool:
        """
        Returns whether this format is able to load a given world.

        :param directory: The path to the root of the world to load.
        :return: True if the world can be loaded by this format, False otherwise.
        """
        if not check_all_exist(directory, "level.dat"):
            return False

        try:
            level_dat_root = load_leveldat(directory)
        except:
            return False

        if "Data" not in level_dat_root:
            return False

        if "FML" in level_dat_root:
            return True

        return False

    @property
    def game_version_string(self) -> str:
        try:
            return f'Java Forge {self.root_tag["Data"]["Version"]["Name"].value}'
        except Exception:
            return f"Java Forge Unknown Version"


export = AnvilForgeFormat



from typing import List, Union, Type
import numpy

import amulet_nbt

from amulet.api.block import Block
from amulet.api.entity import Entity
from amulet.api.block_entity import BlockEntity
from amulet.api.registry import BlockManager


def unpack_palette(raw_palette: amulet_nbt.TAG_List) -> List[Block]:
    block_palette = []
    extra_block_map = {}
    for block_index, block_nbt in enumerate(raw_palette):
        block_nbt: amulet_nbt.TAG_Compound
        block_namespace = block_nbt["namespace"].value
        block_basename = block_nbt["blockname"].value
        block = Block(
            namespace=block_namespace,
            base_name=block_basename,
            properties=block_nbt["properties"].value,
        )

        if block_nbt["extra_blocks"].value:
            extra_block_map[block_index] = block_nbt["extra_blocks"].value

        block_palette.append(block)

    for block_index, extra_blocks in extra_block_map.items():
        extra_block_objects = [block_palette[i.value] for i in extra_blocks]

        resulting_block = block_palette[block_index]
        for extra_block in extra_block_objects:
            resulting_block = resulting_block + extra_block

        block_palette[block_index] = resulting_block
    return block_palette


def parse_entities(entities: amulet_nbt.TAG_List) -> List[Entity]:
    return [
        Entity(
            entity["namespace"].value,
            entity["base_name"].value,
            entity["x"].value,
            entity["y"].value,
            entity["z"].value,
            amulet_nbt.NBTFile(entity["nbt"]),
        )
        for entity in entities
    ]


def parse_block_entities(block_entities: amulet_nbt.TAG_List) -> List[BlockEntity]:
    return [
        BlockEntity(
            block_entity["namespace"].value,
            block_entity["base_name"].value,
            block_entity["x"].value,
            block_entity["y"].value,
            block_entity["z"].value,
            amulet_nbt.NBTFile(block_entity["nbt"]),
        )
        for block_entity in block_entities
    ]


def generate_block_entry(
    block: Block, palette_len, extra_blocks
) -> amulet_nbt.TAG_Compound:
    return amulet_nbt.TAG_Compound(
        {
            "namespace": amulet_nbt.TAG_String(block.namespace),
            "blockname": amulet_nbt.TAG_String(block.base_name),
            "properties": amulet_nbt.TAG_Compound(block.properties),
            "extra_blocks": amulet_nbt.TAG_List(
                [
                    amulet_nbt.TAG_Int(palette_len + extra_blocks.index(_extra_block))
                    for _extra_block in block.extra_blocks
                ]
            ),
        }
    )


def serialise_entities(entities: List[Entity]) -> amulet_nbt.TAG_List:
    return amulet_nbt.TAG_List(
        [
            amulet_nbt.TAG_Compound(
                {
                    "namespace": amulet_nbt.TAG_String(entity.namespace),
                    "base_name": amulet_nbt.TAG_String(entity.base_name),
                    "x": amulet_nbt.TAG_Double(entity.x),
                    "y": amulet_nbt.TAG_Double(entity.y),
                    "z": amulet_nbt.TAG_Double(entity.z),
                    "nbt": entity.nbt.value,
                }
            )
            for entity in entities
        ]
    )


def serialise_block_entities(
    block_entities: List[BlockEntity],
) -> amulet_nbt.TAG_List:
    return amulet_nbt.TAG_List(
        [
            amulet_nbt.TAG_Compound(
                {
                    "namespace": amulet_nbt.TAG_String(block_entity.namespace),
                    "base_name": amulet_nbt.TAG_String(block_entity.base_name),
                    "x": amulet_nbt.TAG_Int(block_entity.x),
                    "y": amulet_nbt.TAG_Int(block_entity.y),
                    "z": amulet_nbt.TAG_Int(block_entity.z),
                    "nbt": block_entity.nbt.value,
                }
            )
            for block_entity in block_entities
        ]
    )


def find_fitting_array_type(
    array: numpy.ndarray,
) -> Union[
    Type[amulet_nbt.TAG_Int_Array],
    Type[amulet_nbt.TAG_Byte_Array],
    Type[amulet_nbt.TAG_Long_Array],
]:
    max_element = array.max(initial=0)

    if max_element <= 127:
        return amulet_nbt.TAG_Byte_Array
    elif max_element <= 2_147_483_647:
        return amulet_nbt.TAG_Int_Array
    else:
        return amulet_nbt.TAG_Long_Array


def pack_palette(palette: BlockManager) -> amulet_nbt.TAG_List:
    block_palette_nbt = amulet_nbt.TAG_List()
    extra_blocks = set()

    for block in palette.blocks():
        if len(block.extra_blocks) > 0:
            extra_blocks.update(block.extra_blocks)

    extra_blocks = list(extra_blocks)

    palette_len = len(palette)
    for block_entry in palette.blocks():
        block_palette_nbt.append(
            generate_block_entry(block_entry, palette_len, extra_blocks)
        )

    for extra_block in extra_blocks:
        block_palette_nbt.append(
            generate_block_entry(extra_block, palette_len, extra_blocks)
        )
    return block_palette_nbt



from __future__ import annotations

from .leveldb_17 import (
    LevelDB17Interface,
)


class LevelDB18Interface(LevelDB17Interface):
    def __init__(self):
        super().__init__()

        self.features["chunk_version"] = 18


export = LevelDB18Interface



from typing import Union, Tuple
import numpy

# from nptyping import NDArray

FloatTriplet = Tuple[float, float, float]

Dimension = str
RegionCoordinates = Tuple[int, int]
ChunkCoordinates = Tuple[int, int]
SubChunkCoordinates = Tuple[int, int, int]
DimensionCoordinates = Tuple[Dimension, int, int]

BlockCoordinates = Tuple[int, int, int]
BlockCoordinatesNDArray = numpy.ndarray  # NDArray[(3, ), int]
BlockCoordinatesAny = Union[BlockCoordinates, BlockCoordinatesNDArray]
PointCoordinates = FloatTriplet
PointCoordinatesNDArray = numpy.ndarray  # NDArray[(3, ), numpy.float]
PointCoordinatesAny = Union[PointCoordinates, PointCoordinatesNDArray]
CoordinatesAny = Union[BlockCoordinatesAny, PointCoordinatesAny]

SubChunkNDArray = numpy.ndarray  # NDArray[(16, 16, 16), numpy.uint32]

BiomeType = str



import os
import warnings
from typing import Any, Generator, List, Tuple, Dict, Optional

from amulet import IMG_DIRECTORY
from amulet.api.data_types import Dimension, PlatformType, ChunkCoordinates
from .format_wrapper import FormatWrapper

missing_world_icon = os.path.abspath(
    os.path.join(IMG_DIRECTORY, "missing_world_icon.png")
)


class WorldFormatWrapper(FormatWrapper):
    _missing_world_icon = missing_world_icon

    def __init__(self, world_path: str):
        if type(self) is WorldFormatWrapper:
            raise Exception(
                "WorldFormatWrapper is not directly usable. One of its subclasses must be used."
            )
        super().__init__(world_path)
        self._world_image_path = missing_world_icon
        self._changed: bool = False

    @property
    def world_name(self) -> str:
        """The name of the world"""
        return "Unknown World"

    @world_name.setter
    def world_name(self, value: str):
        raise NotImplementedError

    @property
    def last_played(self) -> int:
        raise NotImplementedError

    @property
    def game_version_string(self) -> str:
        raise NotImplementedError

    @property
    def world_path(self) -> str:
        """The path to the world directory"""
        warnings.warn(
            "Format.world_path is depreciated and will be removed in the future. Please used WorldFormatWrapper.path instead",
            DeprecationWarning,
        )
        return self._path

    @property
    def world_image_path(self) -> str:
        """The path to the world icon"""
        return self._world_image_path

    @staticmethod
    def is_valid(path: str) -> bool:
        raise NotImplementedError

    @property
    def valid_formats(self) -> Dict[PlatformType, Tuple[bool, bool]]:
        raise NotImplementedError

    @property
    def can_add_dimension(self) -> bool:
        """Can external code register a new dimension.
        If False register_dimension will have no effect."""
        return True

    @property
    def dimensions(self) -> List[Dimension]:
        raise NotImplementedError

    def register_dimension(self, dimension_internal: Any, dimension_name: Dimension):
        raise NotImplementedError

    def _get_interface(self, raw_chunk_data: Optional[Any] = None) -> "Interface":
        from amulet.level.loader import Interfaces

        key = self._get_interface_key(raw_chunk_data)
        return Interfaces.get(key)

    def _get_interface_key(self, raw_chunk_data: Optional[Any] = None) -> Any:
        raise NotImplementedError

    def _create(self, overwrite: bool, **kwargs):
        raise NotImplementedError

    def _open(self):
        raise NotImplementedError

    def _save(self):
        raise NotImplementedError

    def _close(self):
        raise NotImplementedError

    def unload(self):
        raise NotImplementedError

    def all_chunk_coords(
        self, dimension: Dimension
    ) -> Generator[ChunkCoordinates, None, None]:
        raise NotImplementedError

    def has_chunk(self, cx: int, cz: int, dimension: Dimension) -> bool:
        raise NotImplementedError

    def _delete_chunk(self, cx: int, cz: int, dimension: Dimension):
        raise NotImplementedError

    def _put_raw_chunk_data(self, cx: int, cz: int, data: Any, dimension: Dimension):
        raise NotImplementedError

    def _get_raw_chunk_data(self, cx: int, cz: int, dimension: Dimension) -> Any:
        raise NotImplementedError



from typing import Dict, overload, List, Union, Tuple, Generator, Iterable

from amulet.api.data_types import Int, BiomeType
from amulet.api.registry.base_registry import BaseRegistry


class BiomeManager(BaseRegistry):
    def __init__(self, biomes: Iterable[BiomeType] = ()):
        self._index_to_biome: List[BiomeType] = []
        self._biome_to_index: Dict[BiomeType, int] = {}
        for biome in biomes:
            # if a list is given it is assumed that the block palette will be the same size as the list.
            # Ensure that if a value is duplicated it will appear twice in the list
            assert isinstance(biome, str), f"biome must be a string. Got {biome}"
            if biome not in self._biome_to_index:
                self._biome_to_index[biome] = len(self._index_to_biome)
            self._index_to_biome.append(biome)

    def __len__(self):
        return len(self._index_to_biome)

    def __contains__(self, item: Union[int, BiomeType]):
        if isinstance(item, int):
            return item < len(self._index_to_biome)
        elif isinstance(item, str):
            return item in self._biome_to_index
        return False

    def __iter__(self) -> Iterable[BiomeType]:
        yield from self._index_to_biome

    def biomes(self) -> Tuple[BiomeType]:
        return tuple(self._index_to_biome)

    def values(self) -> Tuple[BiomeType]:
        return self.biomes()

    def items(self) -> Generator[Tuple[int, BiomeType], None, None]:
        yield from enumerate(self._index_to_biome)

    @overload
    def __getitem__(self, item: BiomeType) -> int:
        ...

    @overload
    def __getitem__(self, item: Int) -> BiomeType:
        ...

    def __getitem__(self, item):
        """
        If a string is passed to this function, it will return the internal ID/index of the biome.
        If an int is given, this method will return the biome string at that specified index.
        If the key is not present KeyError will be raised.

        :param item: The string or int to get the mapping data of
        :return: An int if a string was supplied, a string if an int was supplied
        """
        try:
            if isinstance(item, str):
                return self._biome_to_index[item]
            return self._index_to_biome[item]

        except (KeyError, IndexError):
            raise KeyError(
                f"There is no {item} in the BiomeManager. "
                f"You might want to use the `get_add_biome` function for the biome before accessing them."
            )

    def get_add_biome(self, biome: BiomeType) -> int:
        """Register the biome and get the runtime id."""
        assert isinstance(biome, str), f"biome must be a string. Got {biome}"
        if biome not in self._biome_to_index:
            self._biome_to_index[biome] = len(self._index_to_biome)
            self._index_to_biome.append(biome)
        return self._biome_to_index[biome]

    def register(self, biome: BiomeType) -> int:
        return self.get_add_biome(biome)



class BaseHistory:
    """The base class for all history related objects"""

    def undo(self):
        """Revert to the previous state."""
        raise NotImplementedError

    def redo(self):
        """Un-revert to the next state."""
        raise NotImplementedError

    def mark_saved(self):
        """Let the class know that the current state has been saved."""
        raise NotImplementedError

    @property
    def changed(self) -> bool:
        """Have there been modifications since the last save."""
        raise NotImplementedError



from typing import Union, Dict, List, Tuple
from amulet import log
import weakref

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from amulet.api.chunk import Chunk

states = {
    "empty": [["j13", "j14"], -1],
    "structure_starts": [["j14"], -0.9],
    "structure_references": [["j14"], -0.8],
    "biomes": [["j14"], -0.7],
    "noise": [["j14"], -0.6],
    "base": [["j13"], -0.5],
    "surface": [["j14"], -0.5],
    "carved": [["j13"], -0.4],
    "carvers": [["j14"], -0.4],
    "liquid_carved": [["j13"], -0.3],
    "liquid_carvers": [["j14"], -0.3],
    "decorated": [["j13"], -0.2],
    "features": [["j14"], -0.2],
    "lighted": [["j13"], -0.1],
    "light": [["j14"], -0.1],
    # 0.0	needs ticked
    # 1.0	needs population
    "mobs_spawned": [["j13"], 1.1],
    "spawn": [["j14"], 1.1],
    "finalized": [["j13"], 1.5],
    "heightmaps": [["j14"], 1.5],
    "fullchunk": [["j13"], 1.9],
    "postprocessed": [["j13"], 2.0],
    "full": [["j14"], 2.0],
    # 2.0	done
}

versions: Dict[str, List[Tuple[float, str]]] = {}
for key, val in states.items():
    for version_ in val[0]:
        versions.setdefault(version_, []).append((val[1], key))
for data in versions.values():
    data.sort(key=lambda val_: val_[0])
    data.reverse()


class Status:
    def __init__(self, parent_chunk: "Chunk"):
        self._parent_chunk = weakref.ref(parent_chunk)
        self._value = 2.0

    @property
    def value(self) -> float:
        return self._value

    @value.setter
    def value(self, value: Union[float, int, str]):
        if isinstance(value, float):
            self._value = value
        elif isinstance(value, int):
            self._value = float(value)
        elif isinstance(value, str) and value in states:
            self._value = states[value][1]
        else:
            log.error(
                f"Unrecognised chunk state {value}. Defaulting to fully generated.\nIf this is a new version report it to the developers."
            )
            self._value = 2.0

    def as_type(self, version: str) -> Union[int, float, str]:
        if version == "float":
            return self._value

        elif version in versions:  # Java 1.13/1.14
            value = next((v for v in versions[version] if v[0] <= self._value), None)
            if value is None:
                value = next((v for v in versions[version] if v[0] <= 2.0), None)
            return value[1]

        elif version == "b":  # Bedrock (0, 1 or 2)
            return int(max(min(2, self._value), 0))



from typing import List, Any

from amulet.api.history.base.history_manager import HistoryManager

SnapshotType = Any


class ContainerHistoryManager(HistoryManager):
    def __init__(self):
        self._snapshots: List[SnapshotType] = []
        self._snapshot_index: int = -1

        # the snapshot that was saved or the save branches off from
        self._last_save_snapshot = -1
        self._branch_save_count = 0  # if the user saves, undoes and does a new operation a save branch will be lost
        # This is the number of changes on that branch

    def _check_snapshot(self, snapshot: SnapshotType):
        raise NotImplementedError

    def _register_snapshot(self, snapshot: SnapshotType) -> bool:
        """Create a snapshot if it contains data.

        :param snapshot: The snapshot data to save.
        :return: True if the snapshot was created, false otherwise.
        """
        self._check_snapshot(snapshot)
        if snapshot:
            if self._last_save_snapshot > self._snapshot_index:
                # if the user has undone changes and made more changes things get a bit messy
                # This fixes the property storing the number of changes since the last save.
                self._branch_save_count += (
                    self._last_save_snapshot - self._snapshot_index
                )
                self._last_save_snapshot = self._snapshot_index
            self._snapshot_index += 1
            # delete all upstream snapshots
            del self._snapshots[self._snapshot_index :]
            self._snapshots.append(snapshot)
            return True
        return False

    def mark_saved(self):
        """Let the class know that the current state has been saved."""
        self._last_save_snapshot = self._snapshot_index
        self._branch_save_count = 0
        self._mark_saved()

    def _mark_saved(self):
        raise NotImplementedError

    @property
    def undo_count(self) -> int:
        return self._snapshot_index + 1

    @property
    def redo_count(self) -> int:
        return len(self._snapshots) - (self._snapshot_index + 1)

    @property
    def unsaved_changes(self) -> int:
        """The number of changes that have been made since the last save"""
        return (
            abs(self._snapshot_index - self._last_save_snapshot)
            + self._branch_save_count
        )

    def undo(self):
        """Undoes the last set of changes to the database"""
        if self.undo_count > 0:
            snapshot = self._snapshots[self._snapshot_index]
            self._undo(snapshot)
            self._snapshot_index -= 1

    def _undo(self, snapshot: SnapshotType):
        raise NotImplementedError

    def redo(self):
        """Redoes the last set of changes to the database"""
        if self.redo_count > 0:
            self._snapshot_index += 1
            snapshot = self._snapshots[self._snapshot_index]
            self._redo(snapshot)

    def _redo(self, snapshot: SnapshotType):
        raise NotImplementedError

    @property
    def changed(self) -> bool:
        return bool(self.unsaved_changes)

    def create_undo_point(self) -> bool:
        raise NotImplementedError

    def restore_last_undo_point(self):
        raise NotImplementedError

    def purge(self):
        """Unload all cached data. Effectively returns the class to its starting state."""
        self._snapshots.clear()
        self._snapshot_index: int = -1
        self._last_save_snapshot = -1
        self._branch_save_count = 0



from __future__ import annotations

from .leveldb_16 import (
    LevelDB16Interface,
)


class LevelDB17Interface(LevelDB16Interface):
    def __init__(self):
        super().__init__()

        self.features["chunk_version"] = 17


export = LevelDB17Interface



from __future__ import annotations

import os
from typing import Dict, Generator
import re

import amulet_nbt as nbt

from amulet.utils import world_utils
from amulet.api.errors import ChunkDoesNotExist
from amulet.api.data_types import (
    ChunkCoordinates,
    RegionCoordinates,
)
from .region import AnvilRegion

InternalDimension = str


class AnvilDimensionManager:
    level_regex = re.compile(r"DIM(?P<level>-?\d+)")

    def __init__(self, directory: str, mcc=False):
        self._directory = directory
        self._regions: Dict[RegionCoordinates, AnvilRegion] = {}
        self._mcc = mcc

        # shallow load all of the existing region classes
        region_dir = os.path.join(self._directory, "region")
        if os.path.isdir(region_dir):
            for region_file_name in os.listdir(region_dir):
                rx, rz = AnvilRegion.get_coords(region_file_name)
                if rx is None:
                    continue
                self._regions[(rx, rz)] = AnvilRegion(
                    os.path.join(self._directory, "region", region_file_name),
                    mcc=self._mcc,
                )

    def all_chunk_coords(self) -> Generator[ChunkCoordinates, None, None]:
        for region in self._regions.values():
            yield from region.all_chunk_coords()

    def has_chunk(self, cx: int, cz: int) -> bool:
        key = world_utils.chunk_coords_to_region_coords(cx, cz)
        return key in self._regions and self._regions[key].has_chunk(
            cx & 0x1F, cz & 0x1F
        )

    def save(self, unload=True):
        # use put_chunk_data to actually upload modified chunks
        # run this to push those changes to disk

        os.makedirs(self._directory, exist_ok=True)
        for region in self._regions.values():
            region.save()
            if unload:
                region.unload()

    def close(self):
        pass

    def unload(self):
        for region in self._regions.values():
            region.unload()

    def get_chunk_data(self, cx: int, cz: int) -> nbt.NBTFile:
        """Get an NBTFile of a chunk from the database.
        Will raise ChunkDoesNotExist if the region or chunk does not exist
        """
        # get the region key
        return self._get_region(cx, cz).get_chunk_data(cx & 0x1F, cz & 0x1F)

    def _get_region(self, cx: int, cz: int, create=False) -> AnvilRegion:
        key = rx, rz = world_utils.chunk_coords_to_region_coords(cx, cz)
        if key in self._regions:
            return self._regions[key]

        if create:
            file_path = os.path.join(self._directory, "region", f"r.{rx}.{rz}.mca")
            self._regions[key] = AnvilRegion(file_path, True, mcc=self._mcc)
        else:
            raise ChunkDoesNotExist

        return self._regions[key]

    def put_chunk_data(self, cx: int, cz: int, data: nbt.NBTFile):
        """pass data to the region file class"""
        # get the region key
        self._get_region(cx, cz, create=True).put_chunk_data(cx & 0x1F, cz & 0x1F, data)

    def delete_chunk(self, cx: int, cz: int):
        try:
            self._get_region(cx, cz).delete_chunk_data(cx & 0x1F, cz & 0x1F)
        except ChunkDoesNotExist:
            pass



from __future__ import annotations

from .anvil_1467 import (
    Anvil1467Interface,
)


class Anvil1484Interface(Anvil1467Interface):
    def __init__(self):
        Anvil1467Interface.__init__(self)
        self.features["height_map"] = "C|V2"

    @staticmethod
    def minor_is_valid(key: int):
        return 1484 <= key < 1503


export = Anvil1484Interface



from __future__ import annotations

import os
import struct
import zlib
import gzip
from typing import Tuple, Dict, Union, Generator, Optional, BinaryIO
import numpy
import time
import re
import threading

import amulet_nbt as nbt

from amulet.utils import world_utils
from amulet.api.errors import ChunkDoesNotExist, ChunkLoadError
from amulet.api.data_types import (
    ChunkCoordinates,
)

InternalDimension = str


class AnvilRegion:
    region_regex = re.compile(r"r\.(?P<rx>-?\d+)\.(?P<rz>-?\d+)\.mca")

    @staticmethod
    def get_coords(file_path: str) -> Tuple[Union[int, None], Union[int, None]]:
        file_path = os.path.basename(file_path)
        match = AnvilRegion.region_regex.fullmatch(file_path)
        if match is None:
            return None, None
        return int(match.group("rx")), int(match.group("rz"))

    def __init__(self, file_path: str, create=False, mcc=False):
        """
        A class wrapper for a region file
        :param file_path: The file path of the region file
        :param create: bool - if true will create the region from scratch. If false will try loading from disk
        """
        self._file_path = file_path
        self.rx, self.rz = self.get_coords(file_path)
        self._mcc = mcc  # create mcc file if the chunk is greater than 1MiB

        # [dirty, mod_time, data_length, data]  feel free to extend if you want to implement modifying in place and defragging
        # Value is Tuple[save time, compressed bytes]. # The latter can be None if the region file has been unloaded to track which chunks exist.
        # If a chunk does not exist on disk it simply won't exist in this dictionary
        self._chunks: Dict[ChunkCoordinates, Tuple[int, Optional[bytes]]] = {}

        # committed chunks stores chunks that have been committed but not saved to disk. Value is Tuple[save time, compressed bytes].
        # The latter can be None in which case the chunk has been deleted
        self._committed_chunks: Dict[ChunkCoordinates, Tuple[int, Optional[bytes]]] = {}

        self._lock = threading.Lock()
        if create:
            # create the region from scratch.
            self._loaded = True
        else:
            # mark the region to be loaded when needed
            self._loaded = False
            # shallow load the data
            file_size = os.path.getsize(self._file_path)
            if file_size > 4096 * 2:
                with open(self._file_path, "rb") as fp:
                    offsets = numpy.fromfile(fp, dtype=">u4", count=1024).reshape(
                        32, 32
                    )
                    for x in range(32):
                        for z in range(32):
                            offset = offsets[z, x]
                            if offset != 0:
                                self._chunks[(x, z)] = (0, None)

    def all_chunk_coords(self) -> Generator[ChunkCoordinates, None, None]:
        for (cx, cz), (_, chunk_) in self._committed_chunks.items():
            if chunk_ is not None:
                yield cx + self.rx * 32, cz + self.rz * 32
        for cx, cz in self._chunks.keys():
            if (cx, cz) not in self._committed_chunks:
                yield cx + self.rx * 32, cz + self.rz * 32

    def has_chunk(self, cx: int, cz: int) -> bool:
        if (cx, cz) in self._committed_chunks:
            return self._committed_chunks[(cx, cz)][1] is not None
        else:
            return (cx, cz) in self._chunks

    def _load_with_lock(self):
        with self._lock:
            self._load()

    def _load(self):
        """Load the data if it is not loaded.
        Note this method on its own is not thread safe. Use _load_with_lock or implement locks yourself."""
        if not self._loaded:
            mode = "rb+" if os.path.isfile(self._file_path) else "wb"
            with open(self._file_path, mode) as fp:
                fp: BinaryIO
                # check that the file is a multiple of 4096 bytes and extend if not
                # TODO: perhaps rewrite this in a way that is more readable
                file_size = os.path.getsize(self._file_path)
                if file_size & 0xFFF:
                    file_size = (file_size | 0xFFF) + 1
                    fp.truncate(file_size)

                # if the length of the region file is 0 extend it to 8KiB TODO (perhaps have some error)
                if file_size < world_utils.SECTOR_BYTES * 2:
                    file_size = world_utils.SECTOR_BYTES * 2
                    fp.truncate(file_size)

                # read the file and populate the internal database
                # self._file_size = file_size

                fp.seek(0)

                # offsets = fp.read(world_utils.SECTOR_BYTES)
                # mod_times = fp.read(world_utils.SECTOR_BYTES)

                # self._free_sectors = free_sectors = numpy.full(
                #     file_size // world_utils.SECTOR_BYTES, True, bool
                # )
                # self._free_sectors[0:2] = False, False

                # the first array is made of 3 byte sector offset and 1 byte sector count
                sectors = (
                    numpy.fromfile(fp, dtype=">u4", count=1024).reshape(32, 32) >> 8
                )
                mod_times = numpy.fromfile(fp, dtype=">u4", count=1024).reshape(32, 32)

                # for offset in offsets:
                #     sector = offset >> 8
                #     count = offset & 0xFF
                #
                #     for i in range(sector, sector + count):
                #         if i >= len(free_sectors):
                #             return False
                #
                #         free_sectors[i] = False

                for cx in range(32):
                    for cz in range(32):
                        sector = sectors[cz, cx]
                        if sector:
                            fp.seek(world_utils.SECTOR_BYTES * sector)
                            # read int value and then read that amount of data
                            buffer_size_bytes: bytes = fp.read(4)
                            buffer_size = struct.unpack(">I", buffer_size_bytes)[0]
                            buffer: bytes = fp.read(buffer_size)

                            if buffer:
                                if buffer[0] & 128:  # if the "external" bit is set
                                    if self._mcc:
                                        mcc_path = os.path.join(
                                            os.path.dirname(self._file_path),
                                            f"c.{cx + self.rx * 32}.{cz + self.rz * 32}.mcc",
                                        )
                                        if os.path.isfile(mcc_path):
                                            with open(mcc_path, "rb") as f:
                                                buffer = (
                                                    bytes([buffer[0] & 127]) + f.read()
                                                )
                                        else:
                                            # the external flag was set but the external file cannot be found. Continue as if the chunk does not exist.
                                            continue
                                    else:
                                        # External bit set but this version cannot handle mcc files. Continue as if the chunk does not exist.
                                        continue

                                self._chunks[(cx, cz)] = (
                                    mod_times[cz, cx],
                                    buffer,
                                )

            self._loaded = True

    def unload(self):
        """Unload the data if it is not being used."""
        if not self._lock.locked():
            # if the data is not being used
            with self._lock:
                # acquire the lock and unload the data
                for key in self._chunks.keys():
                    self._chunks[key] = (0, None)
                self._loaded = False

    def save(self):
        if self._committed_chunks:
            with self._lock:
                # acquire the lock so that it cannot get unloaded while saving.
                self._load()
                if self._mcc:
                    mcc_chunks = {(cx, cz) for cx in range(32) for cz in range(32)}
                else:
                    mcc_chunks = set()
                for key, val in self._committed_chunks.items():
                    if val[1] is not None:
                        if self._mcc or len(val[1]) <= 2 ** 20 - 4:
                            self._chunks[key] = val
                    elif key in self._chunks:
                        del self._chunks[key]
                self._committed_chunks.clear()
                offsets = numpy.zeros(1024, dtype=">u4")
                mod_times = numpy.zeros(1024, dtype=">u4")
                offset = 2
                data = []
                for (cx, cz), (mod_time, buffer) in self._chunks.items():
                    if buffer is not None:
                        index = cx + (cz << 5)
                        buffer_size = len(buffer)
                        if (
                            buffer_size > 2 ** 20 - 4
                        ):  # if mcc is false the chunks that are too large should have already been removed.
                            mcc_chunks.remove((cx, cz))
                            with open(
                                os.path.join(
                                    os.path.dirname(self._file_path),
                                    f"c.{cx + self.rx * 32}.{cz + self.rz * 32}.mcc",
                                ),
                                "wb",
                            ) as f:
                                f.write(buffer[1:])
                            buffer = bytes([buffer[0] | 128])  # set the external flag
                            buffer_size = 1

                        sector_count = ((buffer_size + 4 | 0xFFF) + 1) >> 12
                        offsets[index] = (offset << 8) + sector_count
                        mod_times[index] = mod_time
                        data.append(
                            struct.pack(">I", buffer_size)
                            + buffer
                            + b"\x00" * ((sector_count << 12) - buffer_size - 4)
                        )
                        offset += sector_count
                os.makedirs(os.path.dirname(self._file_path), exist_ok=True)
                with open(self._file_path, "wb") as fp:
                    fp.write(
                        struct.pack(">1024I", *offsets)
                    )  # there is probably a prettier way of doing this
                    fp.write(
                        struct.pack(">1024I", *mod_times)
                    )  # but I could not work it out with Numpy
                    fp.write(b"".join(data))

                # remove orphaned mcc files
                for cx, cz in mcc_chunks:
                    mcc_path = os.path.join(
                        os.path.dirname(self._file_path),
                        f"c.{cx + self.rx * 32}.{cz + self.rz * 32}.mcc",
                    )
                    if os.path.isfile(mcc_path):
                        os.remove(mcc_path)

    def get_chunk_data(self, cx: int, cz: int) -> nbt.NBTFile:
        """Get chunk data. Coords are in region space."""
        if (cx, cz) in self._committed_chunks:
            # if the chunk exists in the committed but unsaved chunks return that
            data = self._committed_chunks[(cx, cz)][1]
            if data is not None:
                compress_type, data = data[0], data[1:]
                return self._decompress(compress_type, data)
        elif (cx, cz) in self._chunks:
            # otherwise if the chunk exists in the main database return that
            with self._lock:
                self._load()
                data: bytes = self._chunks[(cx, cz)][1]
                compress_type, data = data[0], data[1:]
                if data:
                    return self._decompress(compress_type, data)

        raise ChunkDoesNotExist

    def put_chunk_data(self, cx: int, cz: int, data: nbt.NBTFile):
        """compress the data and put it in the class database"""
        bytes_data = self._compress(data)
        self._committed_chunks[(cx, cz)] = (int(time.time()), bytes_data)

    def delete_chunk_data(self, cx: int, cz: int):
        self._committed_chunks[(cx, cz)] = (0, None)

    @staticmethod
    def _compress(data: nbt.NBTFile) -> bytes:
        """Convert an NBTFile into a compressed bytes object"""
        data = data.save_to(compressed=False)
        return b"\x02" + zlib.compress(data)

    @staticmethod
    def _decompress(compress_type: int, data: bytes) -> nbt.NBTFile:
        """Convert a bytes object into an NBTFile"""
        if compress_type == world_utils.VERSION_GZIP:
            return nbt.load(gzip.decompress(data), compressed=False)
        elif compress_type == world_utils.VERSION_DEFLATE:
            return nbt.load(zlib.decompress(data), compressed=False)
        raise ChunkLoadError(f"Invalid compression type {compress_type}")



from __future__ import annotations

from .anvil_1466 import (
    Anvil1466Interface,
)


class Anvil1467Interface(Anvil1466Interface):
    def __init__(self):
        Anvil1466Interface.__init__(self)
        self.features["biomes"] = "256IA"

    @staticmethod
    def minor_is_valid(key: int):
        return 1467 <= key < 1484


export = Anvil1467Interface



from __future__ import annotations

from .leveldb_2 import (
    LevelDB2Interface,
)


class LevelDB3Interface(LevelDB2Interface):
    def __init__(self):
        LevelDB2Interface.__init__(self)

        self.features["chunk_version"] = 3

        self.features["terrain"] = "2farray"


export = LevelDB3Interface



from typing import TYPE_CHECKING, Any, Tuple, Union
import numpy

from amulet.api.wrapper import Interface, EntityIDType, EntityCoordType
from amulet.api.chunk import Chunk
from amulet.api.selection import SelectionBox
from amulet.api.data_types import AnyNDArray
from amulet.level.loader import Translators
from amulet.api.block import Block
from .chunk import SpongeSchemChunk

if TYPE_CHECKING:
    from amulet.api.wrapper import Translator
    from PyMCTranslate import TranslationManager


class SpongeSchemInterface(Interface):
    _entity_id_type = EntityIDType.namespace_str_Id
    _entity_coord_type = EntityCoordType.Pos_list_double
    _block_entity_id_type = EntityIDType.namespace_str_Id
    _block_entity_coord_type = EntityCoordType.Pos_array_int

    def is_valid(self, key: Tuple) -> bool:
        return True

    def decode(
        self, cx: int, cz: int, section: SpongeSchemChunk
    ) -> Tuple["Chunk", AnyNDArray]:
        palette = numpy.empty(len(section.palette) + 1, dtype=object)
        palette[0] = Block(namespace="minecraft", base_name="air")
        palette[1:] = section.palette[:]

        chunk = Chunk(cx, cz)
        box = section.selection.create_moved_box((cx * 16, 0, cz * 16), subtract=True)
        chunk.blocks[box.slice] = section.blocks + 1
        for b in section.block_entities:
            b = self._decode_block_entity(
                b, self._block_entity_id_type, self._block_entity_coord_type
            )
            if b is not None:
                chunk.block_entities.insert(b)
        for e in section.entities:
            e = self._decode_entity(
                e, self._block_entity_id_type, self._block_entity_coord_type
            )
            if e is not None:
                chunk.entities.append(e)

        return chunk, palette

    def encode(
        self,
        chunk: "Chunk",
        palette: AnyNDArray,
        max_world_version: Tuple[str, Union[int, Tuple[int, int, int]]],
        box: SelectionBox = None,
    ) -> SpongeSchemChunk:
        entities = []
        for e in chunk.entities:
            if e.location in box:
                entities.append(
                    self._encode_entity(
                        e, self._entity_id_type, self._entity_coord_type
                    ).value
                )
        block_entities = []
        for e in chunk.block_entities:
            if e.location in box:
                block_entities.append(
                    self._encode_block_entity(
                        e, self._block_entity_id_type, self._block_entity_coord_type
                    ).value
                )

        slices = box.create_moved_box(
            (chunk.cx * 16, 0, chunk.cz * 16), subtract=True
        ).slice

        return SpongeSchemChunk(
            box,
            numpy.asarray(chunk.blocks[slices]),
            palette.copy(),
            block_entities,
            entities,
        )

    def get_translator(
        self,
        max_world_version: Tuple[str, int],
        data: Any = None,
        translation_manager: "TranslationManager" = None,
    ) -> Tuple["Translator", Union[int, Tuple[int, int, int]]]:
        platform, version_number = max_world_version
        if platform != "java":
            raise ValueError("Platform must be java")
        version = translation_manager.get_version(platform, version_number)
        version_number = version.data_version
        return Translators.get((platform, version_number)), version_number



from typing import Optional, Dict, Union, Tuple, Type, TYPE_CHECKING
import numpy

import copy

from .util import get_sanitised_sliced_array_size
from .data_types import DtypeType, UnpackedSlicesType

if TYPE_CHECKING:
    from .unbounded_partial_3d_array import UnboundedPartial3DArray


class BasePartial3DArray:
    """Do not use this class directly. Use UnboundedPartial3DArray or BoundedPartial3DArray"""

    def __init__(
        self,
        dtype: DtypeType,
        default_value: Union[int, bool],
        section_shape: Tuple[int, int, int],
        start: Tuple[Optional[int], Optional[int], Optional[int]],
        stop: Tuple[Optional[int], Optional[int], Optional[int]],
        step: Tuple[Optional[int], Optional[int], Optional[int]],
        parent_array: Optional["UnboundedPartial3DArray"] = None,
        sections: Optional[Dict[int, numpy.ndarray]] = None,
    ):
        self._dtype = dtype
        self._default_value: Union[int, bool] = default_value
        self._section_shape = section_shape
        self._parent_array = parent_array

        self._size_x = None
        self._size_y = None
        self._size_z = None

        self._start_x, self._stop_x, self._step_x = start[0], stop[0], step[0]
        self._start_y, self._stop_y = start[1], stop[1]
        assert (self._start_y is None and self._stop_y is None) or (
            isinstance(self._start_y, int) and isinstance(self._stop_y, int)
        ), "y start and stop must both be None or ints"
        self._step_y = 1 if step[1] is None else step[1]
        if self.size_y == 0:
            self._stop_y = self._start_y
        self._start_z, self._stop_z, self._step_z = start[2], stop[2], step[2]

        if parent_array is None:
            # populate from sections
            self._sections: Dict[int, numpy.ndarray] = sections or {}
            for key, section in self._sections.items():
                assert isinstance(key, int), "All keys must be ints"
                assert (
                    section.shape == self._section_shape
                ), "The size of all sections must be equal to the section_shape."
                assert (
                    section.dtype == self._dtype
                ), "The given dtype does not match the arrays given."

        elif isinstance(parent_array, BasePartial3DArray):
            # populate from the array
            assert (
                parent_array.section_shape == self._section_shape
            ), "The parent section shape must match the given section_shape."
            assert (
                parent_array.dtype == self._dtype
            ), "The parent dtype must match the given dtype"
            self._sections = parent_array._sections

        else:
            raise Exception(
                f"{parent_array.__class__.__name__}({parent_array}) is not a valid input type for parent_array"
            )

    @property
    def start_x(self) -> int:
        return self._start_x

    @property
    def start_y(self) -> int:
        return self._start_y

    @property
    def start_z(self) -> int:
        return self._start_z

    @property
    def start(self) -> Tuple[int, int, int]:
        return self.start_x, self.start_y, self.start_z

    @property
    def stop_x(self) -> int:
        return self._stop_x

    @property
    def stop_y(self) -> int:
        return self._stop_y

    @property
    def stop_z(self) -> int:
        return self._stop_z

    @property
    def stop(self) -> Tuple[int, int, int]:
        return self.stop_x, self.stop_y, self.stop_z

    @property
    def step_x(self) -> int:
        return self._step_x

    @property
    def step_y(self) -> int:
        return self._step_y

    @property
    def step_z(self) -> int:
        return self._step_z

    @property
    def step(self) -> Tuple[int, int, int]:
        return self.step_x, self.step_y, self.step_z

    @property
    def slice_x(self) -> slice:
        return slice(self.start_x, self.stop_x, self.step_x)

    @property
    def slice_y(self) -> slice:
        return slice(self.start_y, self.stop_y, self.step_y)

    @property
    def slice_z(self) -> slice:
        return slice(self.start_z, self.stop_z, self.step_z)

    @property
    def slices_tuple(self) -> UnpackedSlicesType:
        return (
            (self.start_x, self.stop_x, self.step_x),
            (self.start_y, self.stop_y, self.step_y),
            (self.start_z, self.stop_z, self.step_z),
        )

    @property
    def size_x(self) -> int:
        if self._size_x is None:
            self._size_x = get_sanitised_sliced_array_size(
                self.start_x, self.stop_x, self.step_x
            )
        return self._size_x

    @property
    def size_y(self) -> Union[int, float]:
        if self._size_y is None:
            self._size_y = get_sanitised_sliced_array_size(
                self.start_y, self.stop_y, self.step_y
            )
        return self._size_y

    @property
    def size_z(self) -> int:
        if self._size_z is None:
            self._size_z = get_sanitised_sliced_array_size(
                self.start_z, self.stop_z, self.step_z
            )
        return self._size_z

    @property
    def shape(self) -> Tuple[int, Union[int, float], int]:
        return self.size_x, self.size_y, self.size_z

    @property
    def dtype(self) -> Type[numpy.dtype]:
        return self._dtype

    @property
    def default_value(self) -> Union[int, bool]:
        """The default value to populate undefined sections with."""
        return self._default_value

    @property
    def section_shape(self) -> Tuple[int, int, int]:
        return self._section_shape

    def _section_index(self, y: int) -> Tuple[int, int]:
        """Get the section index and location within the section of an absolute y coordinate"""
        return y // self.section_shape[1], y % self.section_shape[1]

    def __array__(self, dtype=None):
        raise NotImplementedError

    def __contains__(self, item: int):
        return item in self._sections

    def __iter__(self):
        raise Exception(
            "Please use sections method if this is what you are trying to achieve"
        )

    def __getitem__(self, item):
        raise NotImplementedError

    def __setitem__(self, key, value):
        raise NotImplementedError

    def copy(self):
        return copy.copy(self)

    def __copy__(self):
        cls = self.__class__
        result = cls.__new__(cls)
        result.__dict__.update(self.__dict__)
        return result

    def __deepcopy__(self, memodict=None):
        cls = self.__class__
        result = cls.__new__(cls)
        memodict[id(self)] = result
        for k, v in self.__dict__.items():
            setattr(result, k, copy.deepcopy(v, memodict))
        return result



from typing import TYPE_CHECKING, Tuple, Generator, Optional
import numpy

from amulet.api.data_types import Dimension, BlockCoordinates, FloatTriplet
from amulet.api.selection import SelectionGroup, SelectionBox
from amulet.api.block import Block, UniversalAirBlock
from amulet.api.errors import ChunkDoesNotExist, ChunkLoadError
from amulet.api.chunk import Chunk
from amulet.api.registry import BlockManager
from amulet.utils.matrix import transform_matrix, displacement_matrix
import amulet.api.level

if TYPE_CHECKING:
    from .base_level import BaseLevel


def gen_paste_blocks(
    block_palette: BlockManager, skip_blocks: Tuple[Block, ...]
) -> numpy.ndarray:
    """Create a numpy array of all the blocks which should be pasted.

    :param block_palette: The block palette of the chunk.
    :param skip_blocks: Blocks to not copy if they match exactly.
    :return:
    """
    return numpy.vectorize(lambda b: b not in skip_blocks)(block_palette.blocks())


def clone(
    src_structure: "BaseLevel",
    src_dimension: Dimension,
    src_selection: SelectionGroup,
    dst_structure: "BaseLevel",
    dst_dimension: Dimension,
    dst_selection_bounds: SelectionGroup,
    location: BlockCoordinates,
    scale: FloatTriplet = (1.0, 1.0, 1.0),
    rotation: FloatTriplet = (0.0, 0.0, 0.0),
    include_blocks: bool = True,
    include_entities: bool = True,
    skip_blocks: Tuple[Block, ...] = (),
    copy_chunk_not_exist: bool = False,
) -> Generator[float, None, None]:
    """Clone the source object data into the destination object with an optional transform.
    The src and dst can be the same object.
    Note this command may change in the future. Refer to all keyword arguments via the keyword.
    :param src_structure: The source structure to paste into the destination structure.
    :param src_dimension: The dimension of the source structure to use.
    :param src_selection: The area of the source structure to copy.
    :param dst_structure: The destination structure to paste into.
    :param dst_dimension: The dimension of the destination structure to use.
    :param dst_selection_bounds: The area of the destination structure that can be modified.
    :param location: The location where the centre of the `src_structure` will be in the `dst_structure`
    :param scale: The scale in the x, y and z axis. These can be negative to mirror.
    :param rotation: The rotation in degrees around each of the axis.
    :param include_blocks: Include blocks from the `src_structure`.
    :param include_entities: Include entities from the `src_structure`.
    :param skip_blocks: If a block matches a block in this list it will not be copied.
    :param copy_chunk_not_exist: If a chunk does not exist in the source should it be copied over as air. Always False where `src_structure` is a World.
    :return: A generator of floats from 0 to 1 with the progress of the paste operation.
    """
    location = tuple(location)
    if include_blocks or include_entities:
        # we actually have to do something
        if isinstance(src_structure, amulet.api.level.World):
            copy_chunk_not_exist = False

        # TODO: look into if this can be a float so it will always be the exact middle
        rotation_point: numpy.ndarray = (
            (src_selection.max + src_selection.min) // 2
        ).astype(int)

        if src_structure is dst_structure and src_dimension == dst_dimension:
            # copying from an object to itself in the same dimension.
            # if the selections do not overlap this can be achieved directly
            # if they do overlap the selection will first need extracting
            # TODO: implement the above
            if (
                tuple(rotation_point) == location
                and scale == (1.0, 1.0, 1.0)
                and rotation == (0.0, 0.0, 0.0)
            ):
                # The src_object was pasted into itself at the same location. Nothing will change so do nothing.
                return
            src_structure = src_structure.extract_structure(
                src_selection, src_dimension
            )
            src_dimension = src_structure.dimensions[0]

        src_structure: "BaseLevel"

        # TODO: I don't know if this is feasible for large boxes: get the intersection of the source and destination selections and iterate over that to minimise work
        if any(rotation) or any(s != 1 for s in scale):
            rotation_radians = tuple(numpy.radians(rotation))
            transform = numpy.matmul(
                transform_matrix(scale, rotation_radians, location),
                displacement_matrix(*-rotation_point),
            )
            inverse_transform = numpy.linalg.inv(transform)

            dst_selection = (
                src_selection.transform((1, 1, 1), (0, 0, 0), tuple(-rotation_point))
                .transform(scale, rotation_radians, location)
                .intersection(dst_selection_bounds)
            )

            volume = dst_selection.volume
            index = 0

            last_src_cx: Optional[int] = None
            last_src_cz: Optional[int] = None
            src_chunk: Optional[
                Chunk
            ] = None  # None here means the chunk does not exist or failed to load. Treat it as if it was air.
            last_dst_cx: Optional[int] = None
            last_dst_cz: Optional[int] = None
            dst_chunk: Optional[
                Chunk
            ] = None  # None here means the chunk failed to load. Do not modify it.

            # TODO: find a way to do this without doing it block by block
            if include_blocks:
                blocks_to_skip = set(skip_blocks)
                for box in dst_selection.selection_boxes:
                    dst_coords = list(box.blocks())
                    coords_array = numpy.ones((len(dst_coords), 4), dtype=numpy.float)
                    coords_array[:, :3] = dst_coords
                    coords_array[:, :3] += 0.5
                    src_coords = (
                        numpy.floor(numpy.matmul(inverse_transform, coords_array.T))
                        .astype(int)
                        .T[:, :3]
                    )
                    for (dst_x, dst_y, dst_z), (src_x, src_y, src_z) in zip(
                        dst_coords, src_coords
                    ):
                        src_cx, src_cz = (src_x >> 4, src_z >> 4)
                        if (src_cx, src_cz) != (last_src_cx, last_src_cz):
                            last_src_cx = src_cx
                            last_src_cz = src_cz
                            try:
                                src_chunk = src_structure.get_chunk(
                                    src_cx, src_cz, src_dimension
                                )
                            except ChunkLoadError:
                                src_chunk = None

                        dst_cx, dst_cz = (dst_x >> 4, dst_z >> 4)
                        if (dst_cx, dst_cz) != (last_dst_cx, last_dst_cz):
                            last_dst_cx = dst_cx
                            last_dst_cz = dst_cz
                            try:
                                dst_chunk = dst_structure.get_chunk(
                                    dst_cx, dst_cz, dst_dimension
                                )
                            except ChunkDoesNotExist:
                                dst_chunk = dst_structure.create_chunk(
                                    dst_cx, dst_cz, dst_dimension
                                )
                            except ChunkLoadError:
                                dst_chunk = None

                        if dst_chunk is not None:
                            if (dst_x, dst_y, dst_z) in dst_chunk.block_entities:
                                del dst_chunk.block_entities[(dst_x, dst_y, dst_z)]
                            if src_chunk is None:
                                if UniversalAirBlock not in blocks_to_skip:
                                    dst_chunk.blocks[
                                        dst_x % 16, dst_y, dst_z % 16
                                    ] = dst_chunk.block_palette.get_add_block(
                                        UniversalAirBlock
                                    )
                            else:
                                # TODO implement support for individual block rotation
                                block = src_chunk.block_palette[
                                    src_chunk.blocks[src_x % 16, src_y, src_z % 16]
                                ]
                                if block not in blocks_to_skip:
                                    dst_chunk.blocks[
                                        dst_x % 16, dst_y, dst_z % 16
                                    ] = dst_chunk.block_palette.get_add_block(block)
                                    if (
                                        src_x,
                                        src_y,
                                        src_z,
                                    ) in src_chunk.block_entities:
                                        dst_chunk.block_entities[
                                            (dst_x, dst_y, dst_z)
                                        ] = src_chunk.block_entities[
                                            (src_x, src_y, src_z)
                                        ].new_at_location(
                                            dst_x, dst_y, dst_z
                                        )
                            dst_chunk.changed = True

                        yield index / volume
                        index += 1

        else:
            # the transform from the structure location to the world location
            offset = numpy.asarray(location).astype(int) - rotation_point
            moved_min_location = src_selection.min + offset

            iter_count = len(
                list(
                    src_structure.get_moved_coord_slice_box(
                        src_dimension,
                        moved_min_location,
                        src_selection,
                        dst_structure.sub_chunk_size,
                        yield_missing_chunks=copy_chunk_not_exist,
                    )
                )
            )

            count = 0

            for (
                src_chunk,
                src_slices,
                src_box,
                (dst_cx, dst_cz),
                dst_slices,
                dst_box,
            ) in src_structure.get_moved_chunk_slice_box(
                src_dimension,
                moved_min_location,
                src_selection,
                dst_structure.sub_chunk_size,
                create_missing_chunks=copy_chunk_not_exist,
            ):
                src_chunk: Chunk
                src_slices: Tuple[slice, slice, slice]
                src_box: SelectionBox
                dst_cx: int
                dst_cz: int
                dst_slices: Tuple[slice, slice, slice]
                dst_box: SelectionBox

                # load the destination chunk
                try:
                    dst_chunk = dst_structure.get_chunk(dst_cx, dst_cz, dst_dimension)
                except ChunkDoesNotExist:
                    dst_chunk = dst_structure.create_chunk(
                        dst_cx, dst_cz, dst_dimension
                    )
                except ChunkLoadError:
                    count += 1
                    continue

                if include_blocks:
                    # a boolean array specifying if each index should be pasted.
                    paste_blocks = gen_paste_blocks(
                        src_chunk.block_palette, skip_blocks
                    )

                    # create a look up table converting the source block ids to the destination block ids
                    gab = numpy.vectorize(
                        dst_chunk.block_palette.get_add_block, otypes=[numpy.uint32]
                    )
                    lut = gab(src_chunk.block_palette.blocks())

                    # iterate through all block entities in the chunk and work out if the block is going to be overwritten
                    remove_block_entities = []
                    for block_entity_location in dst_chunk.block_entities.keys():
                        if block_entity_location in dst_box:
                            chunk_block_entity_location = (
                                numpy.array(block_entity_location) - offset
                            )
                            chunk_block_entity_location[[0, 2]] %= 16
                            if paste_blocks[
                                src_chunk.blocks[tuple(chunk_block_entity_location)]
                            ]:
                                remove_block_entities.append(block_entity_location)
                    for block_entity_location in remove_block_entities:
                        del dst_chunk.block_entities[block_entity_location]

                    # copy over the source block entities if the source block is supposed to be pasted
                    for (
                        block_entity_location,
                        block_entity,
                    ) in src_chunk.block_entities.items():
                        if block_entity_location in src_box:
                            chunk_block_entity_location = numpy.array(
                                block_entity_location
                            )
                            chunk_block_entity_location[[0, 2]] %= 16
                            if paste_blocks[
                                src_chunk.blocks[tuple(chunk_block_entity_location)]
                            ]:
                                dst_chunk.block_entities.insert(
                                    block_entity.new_at_location(
                                        *offset + block_entity_location
                                    )
                                )

                    mask = paste_blocks[src_chunk.blocks[src_slices]]
                    dst_chunk.blocks[dst_slices][mask] = lut[
                        src_chunk.blocks[src_slices]
                    ][mask]
                    dst_chunk.changed = True

                if include_entities:
                    # TODO: implement pasting entities when we support entities
                    pass

                count += 1
                yield count / iter_count

        yield 1.0



from typing import Optional, Any

from amulet.api.history import Changeable
from ..base import RevisionManager
from amulet.api.history.base.history_manager import HistoryManager
from ..revision_manager import RAMRevisionManager


class ObjectHistoryManager(HistoryManager):
    def __init__(self, original_entry: Changeable):
        super().__init__()
        self._value: Changeable = original_entry
        self._revision_manager = self._create_new_revision_manager(
            self._pack_value(self._value)
        )
        self._snapshots_size: int = 0
        self._snapshot_index: int = -1

        # the snapshot that was saved or the save branches off from
        self._last_save_snapshot = -1
        self._branch_save_count = 0  # if the user saves, undoes and does a new operation a save branch will be lost
        # This is the number of changes on that branch

    @property
    def value(self):
        return self._value

    @staticmethod
    def _create_new_revision_manager(original_entry: Optional[Any]) -> RevisionManager:
        """Create an RevisionManager as desired and populate it with the original entry."""
        return RAMRevisionManager(original_entry)

    def _register_snapshot(self):
        if self._last_save_snapshot > self._snapshot_index:
            # if the user has undone changes and made more changes things get a bit messy
            # This fixes the property storing the number of changes since the last save.
            self._branch_save_count += self._last_save_snapshot - self._snapshot_index
            self._last_save_snapshot = self._snapshot_index
        self._snapshot_index += 1
        self._snapshots_size = min(self._snapshots_size, self._snapshot_index) + 1

    def mark_saved(self):
        """Let the class know that the current state has been saved."""
        self._last_save_snapshot = self._snapshot_index
        self._branch_save_count = 0
        self._revision_manager.mark_saved()

    def purge(self):
        """Unload the cached objects and restore to the starting value."""
        self._revision_manager = self._create_new_revision_manager(
            self._pack_value(self._value)
        )
        self._snapshots_size: int = 0
        self._snapshot_index: int = -1
        self._last_save_snapshot = -1
        self._branch_save_count = 0

    @property
    def undo_count(self) -> int:
        return self._snapshot_index + 1

    @property
    def redo_count(self) -> int:
        return self._snapshots_size - (self._snapshot_index + 1)

    @property
    def unsaved_changes(self) -> int:
        """The number of changes that have been made since the last save"""
        return (
            abs(self._snapshot_index - self._last_save_snapshot)
            + self._branch_save_count
        )

    def undo(self):
        """Undoes the last set of changes to the object"""
        if self.undo_count > 0:
            self._revision_manager.undo()
            self._unpack()
            self._snapshot_index -= 1

    def redo(self):
        """Redoes the last set of changes to the object"""
        if self.redo_count > 0:
            self._snapshot_index += 1
            self._revision_manager.redo()
            self._unpack()

    @property
    def changed(self) -> bool:
        return bool(self.unsaved_changes)

    def create_undo_point(self) -> bool:
        if self._value.changed:
            self._value.changed = False
            self._pack()
            self._register_snapshot()
            return True
        else:
            self._unpack()
            return False

    def restore_last_undo_point(self):
        self._unpack()

    def _unpack(self):
        self._unpack_value(self._revision_manager.get_current_entry())

    def _unpack_value(self, value: Optional[Any]):
        self._value = value

    def _pack(self):
        self._revision_manager.put_new_entry(self._pack_value(self._value))

    def _pack_value(self, value: Optional[Any]) -> Optional[Any]:
        return value



class FormatError(Exception):
    pass


class LoaderNoneMatched(FormatError):
    pass


class FormatLoaderInvalidFormat(FormatError):
    pass


class FormatLoaderMismatched(FormatError):
    pass


class InvalidBlockException(Exception):
    pass


class EntryLoadError(Exception):
    pass


class EntryDoesNotExist(EntryLoadError):
    pass


class ChunkLoadError(EntryLoadError):
    pass


class ChunkSaveError(Exception):
    pass


class LevelDoesNotExist(Exception):
    pass


class ChunkDoesNotExist(EntryDoesNotExist, ChunkLoadError):
    pass


class ObjectReadWriteError(Exception):
    pass


class ObjectReadError(ObjectReadWriteError):
    pass


class ObjectWriteError(ObjectReadError):
    pass



from typing import List, Optional, Any

from .base_history import BaseHistory

StoredEntryType = Optional[Any]


class RevisionManager(BaseHistory):
    """The base API for all passive history manager objects.
    The RevisionManager stores the actual versions of the object to revert to."""

    __slots__ = ("_revisions", "_current_revision_index", "_saved_revision_index")

    def __init__(self, initial_state: StoredEntryType):
        self._revisions: List[StoredEntryType] = []  # the data for each revision
        self._current_revision_index: int = (
            0  # the index into the above for the current data
        )
        self._saved_revision_index: int = (
            0  # the index into the above for the saved version
        )
        self._store_entry(initial_state)

    @property
    def changed(self) -> bool:
        """Have there been modifications since the last save."""
        return self._current_revision_index != self._saved_revision_index

    def put_new_entry(self, entry: StoredEntryType):
        """Add a new entry to the database and increment the index."""
        if len(self._revisions) > self._current_revision_index + 1:
            # if there are upstream revisions delete them
            del self._revisions[self._current_revision_index + 1 :]
        if self._saved_revision_index > self._current_revision_index:
            # we are starting a new branch and the save was on the old branch.
            self._saved_revision_index = -1
        self._store_entry(entry)
        self._current_revision_index += 1

    def _store_entry(self, entry: StoredEntryType):
        """Store the entry data as required."""
        raise NotImplementedError

    def get_current_entry(self):
        """Get the entry at the current revision."""
        raise NotImplementedError

    def undo(self):
        """Decrement the state of the entry to the previous revision."""
        if self._current_revision_index <= 0:
            raise Exception(
                "Cannot undo past revision 0"
            )  # if run there is a bug in the code
        self._current_revision_index -= 1

    def redo(self):
        """Increment the state of the entry to the next revision."""
        if self._current_revision_index >= len(self._revisions):
            raise Exception(
                "Cannot redo past the highest revision"
            )  # if run there is a bug in the code
        self._current_revision_index += 1

    def mark_saved(self):
        """Let the class know that the current state has been saved."""
        self._saved_revision_index = self._current_revision_index

    @property
    def is_deleted(self) -> bool:
        return self._revisions[self._current_revision_index] is None



from __future__ import annotations
from typing import Generator

from ..base_level import BaseLevel
from .void_format_wrapper import VoidFormatWrapper
from amulet.api.chunk import Chunk
from amulet.api.data_types import Dimension
from amulet.api.selection import SelectionGroup, SelectionBox
from amulet.utils.generator import generator_unpacker
import copy


class ImmutableStructure(BaseLevel):
    """
    This is a special class that exists purely to hold chunk data and serialise it to the disk cache.
    There is no level attached to load or save chunks to.
    This class exists purely to store chunks in when copied from a level.
    The original wrapper is mutable so the chunks must be deep copied from it and dropped into here.
    """

    def __init__(
        self,
        temp_dir: str = None,
    ):
        super().__init__("", VoidFormatWrapper(""), temp_dir)
        self._selection = SelectionGroup(
            [
                SelectionBox(
                    (-30_000_000, -30_000_000, -30_000_000),
                    (30_000_000, 30_000_000, 30_000_000),
                )
            ]
        )

    @property
    def selection_bounds(self) -> SelectionGroup:
        return self._selection

    def undo(self):
        pass

    def redo(self):
        pass

    def put_chunk(self, chunk: Chunk, dimension: Dimension):
        super().put_chunk(chunk, dimension)
        self.history_manager.create_undo_point()

    @classmethod
    def from_level(
        cls, level: BaseLevel, selection: SelectionGroup, dimension: Dimension
    ):
        return generator_unpacker(cls.from_level_iter(level, selection, dimension))

    @classmethod
    def from_level_iter(
        cls, level: BaseLevel, selection: SelectionGroup, dimension: Dimension
    ) -> Generator[float, None, ImmutableStructure]:
        """Populate this class with the chunks that intersect the selection."""
        self = cls()
        self._selection = selection.copy()
        dst_dimension = self.dimensions[0]
        count = len(list(level.get_coord_box(dimension, selection)))
        for index, (chunk, _) in enumerate(level.get_chunk_boxes(dimension, selection)):
            self.put_chunk(copy.deepcopy(chunk), dst_dimension)
            yield (index + 1) / count
        return self



from __future__ import annotations

from .leveldb_4 import (
    LevelDB4Interface,
)


class LevelDB5Interface(LevelDB4Interface):
    def __init__(self):
        super().__init__()

        self.features["chunk_version"] = 5


export = LevelDB5Interface



from __future__ import annotations

import numpy
from typing import TYPE_CHECKING

from amulet.level.translators.chunk.bedrock import BaseBedrockTranslator
from amulet.api.data_types import BlockNDArray, AnyNDArray

if TYPE_CHECKING:
    from amulet.api.block import Block
    from PyMCTranslate import Version


class BedrockNumericalTranslator(BaseBedrockTranslator):
    @staticmethod
    def is_valid(key):
        return key[0] == "bedrock" and key[1] < (1, 2, 13)

    def _pack_block_palette(
        self, version: "Version", palette: BlockNDArray
    ) -> AnyNDArray:
        """
        Packs a numpy array of Block objects into an object array containing version number, block ids and block data values.
        :param version:
        :param palette:
        :return: numpy.ndarray[Tuple[Tuple[Optional[VersionNumber], Tuple[int, int]]]]
        """
        palette_ = numpy.empty(len(palette), dtype=object)
        for palette_index, entry in enumerate(palette):
            entry: "Block"
            block_tuple = version.block.block_to_ints(entry)
            if block_tuple is None:
                block_tuple = (0, 0)  # TODO: find some way for the user to specify this
            palette_[palette_index] = ((None, block_tuple),)

        return palette_


export = BedrockNumericalTranslator



from __future__ import annotations

from .leveldb_20 import (
    LevelDB20Interface,
)


class LevelDB21Interface(LevelDB20Interface):
    def __init__(self):
        super().__init__()

        self.features["chunk_version"] = 21


export = LevelDB21Interface



from typing import Union, Tuple, overload, Iterable, Optional, Dict
import numpy
import math

from .base_partial_3d_array import BasePartial3DArray
from .util import sanitise_slice, to_slice, unpack_slice, sanitise_unbounded_slice
from .data_types import DtypeType


class UnboundedPartial3DArray(BasePartial3DArray):
    """This is designed to work similarly to a numpy.ndarray but stores the data in a very different way.
    A numpy.ndarray stores a fixed size continuous array which for large arrays can become unmanageable.
    Sparse arrays allow individual values to exist which can be great where a small set of values are
    defined in a large area but can be less efficient in some cases than a continuous array.

    This class was born out of the need for an array that has a fixed size in the horizontal distances but an
    unlimited height in the vertical distance (both above and below the origin)
    This is achieved by splitting the array into sections of fixed height and storing them sparsely so that only
    the defined sections need to be held in memory.

    This class also implements an API that resembles that of the numpy.ndarray but it is not directly compatible with numpy.
    This class implements methods to access and directly modify the underlying numpy arrays to give finer control.
    """

    def __init__(
        self,
        dtype: DtypeType,
        default_value: Union[int, bool],
        section_shape: Tuple[int, int, int],
        default_section_counts: Tuple[int, int],
        sections: Optional[Dict[int, numpy.ndarray]] = None,
    ):
        super().__init__(
            dtype,
            default_value,
            section_shape,
            (0, None, 0),
            (section_shape[0], None, section_shape[0]),
            (1, 1, 1),
            sections=sections,
        )
        self._default_min_y = -default_section_counts[0] * self.section_shape[1]
        self._default_max_y = default_section_counts[1] * self.section_shape[1]

    def __repr__(self):
        return f"UnboundedPartial3DArray(dtype={self.dtype}, shape={self.shape})"

    @property
    def size_y(self) -> float:
        return math.inf

    @property
    def sections(self) -> Iterable[int]:
        """An iterable of the section indexes that exist"""
        return self._sections.keys()

    def create_section(self, sy: int):
        self._sections[sy] = numpy.full(
            self.section_shape, self.default_value, dtype=self._dtype
        )

    def add_section(self, sy: int, section: numpy.ndarray):
        assert (
            section.shape == self._section_shape
        ), "The size of all sections must be equal to the section_shape."
        if section.dtype != self._dtype:
            section = section.astype(self._dtype)
        self._sections[sy] = section

    def get_section(self, sy: int) -> numpy.ndarray:
        """Get the section ndarray for a given section index.
        :param sy: The section y index
        :return: Numpy array for this section
        :raises:
            KeyError: if no section exists with this index
        """
        if sy not in self._sections:
            self.create_section(sy)
        return self._sections[sy]

    def __setitem__(
        self,
        slices: Tuple[Union[int, slice], Union[int, slice], Union[int, slice]],
        value: Union[int, numpy.integer, numpy.ndarray],
    ):
        if isinstance(value, (int, numpy.integer)) and all(
            isinstance(s, (int, numpy.integer)) for s in slices
        ):
            sy, dy = self._section_index(slices[1])
            self.get_section(sy)[(slices[0], dy, slices[2])] = value
        else:
            self[slices][:, :, :] = value

    @overload
    def __getitem__(self, slices: Tuple[int, int, int]) -> int:
        ...

    @overload
    def __getitem__(
        self, slices: Tuple[Union[int, slice], Union[int, slice], Union[int, slice]]
    ) -> "BoundedPartial3DArray":
        ...

    def __getitem__(self, item):
        if isinstance(item, tuple):
            if len(item) != 3:
                raise KeyError(f"Tuple item must be of length 3, got {len(item)}")
            if all(isinstance(i, (int, numpy.integer)) for i in item):
                sy = item[1] // self.section_shape[1]
                if sy in self:
                    return int(
                        self._sections[sy][
                            (item[0], item[1] % self.section_shape[1], item[2])
                        ]
                    )
                else:
                    return self.default_value

            elif all(isinstance(i, (int, numpy.integer, slice)) for i in item):
                start_x, stop_x, step_x = sanitise_slice(
                    *unpack_slice(to_slice(item[0])), self.size_x
                )
                start_y, stop_y, step_y = sanitise_unbounded_slice(
                    *unpack_slice(to_slice(item[1])),
                    self._default_min_y,
                    self._default_max_y,
                )
                start_z, stop_z, step_z = sanitise_slice(
                    *unpack_slice(to_slice(item[2])), self.size_z
                )

                return BoundedPartial3DArray.from_partial_array(
                    self,
                    (start_x, start_y, start_z),
                    (stop_x, stop_y, stop_z),
                    (step_x, step_y, step_z),
                )
            else:
                raise KeyError(f"Unsupported tuple {item} for getitem")

        else:
            raise KeyError(
                f"{item.__class__.__name__}({item}) is not a supported input for __getitem__"
            )

    def __array__(self, dtype=None):
        return self[:, :, :].__array__(dtype)


from .bounded_partial_3d_array import BoundedPartial3DArray



from __future__ import annotations

from typing import Dict, Iterable, List, Tuple, overload, Generator

from amulet.api.data_types import Int
from amulet.api.block import Block
from .base_registry import BaseRegistry


class BlockManager(BaseRegistry):
    """
    Class to handle the mappings between Block objects and their index-based internal IDs
    """

    def __init__(self, blocks: Iterable[Block] = ()):
        """
        Creates a new BlockManager object
        """
        self._index_to_block: List[Block] = []
        self._block_to_index_map: Dict[Block, int] = {}

        for block in blocks:
            assert isinstance(block, Block), "BlockManager only takes Block objects."
            if block not in self._block_to_index_map:
                self._block_to_index_map[block] = len(self._block_to_index_map)
            self._index_to_block.append(block)

    def __len__(self):
        return len(self._index_to_block)

    def __contains__(self, item: Block) -> bool:
        return item in self._block_to_index_map

    def __iter__(self):
        yield from self._index_to_block

    def blocks(self) -> Tuple[Block]:
        return tuple(self._index_to_block)

    def values(self) -> Tuple[Block]:
        return self.blocks()

    def items(self) -> Generator[Tuple[int, Block], None, None]:
        yield from enumerate(self._index_to_block)

    @overload
    def __getitem__(self, item: Block) -> int:
        ...

    @overload
    def __getitem__(self, item: Int) -> Block:
        ...

    def __getitem__(self, item):
        """
        If a Block object is passed to this function, it will return the internal ID/index of the blockstate.
        If an int is given, this method will return the Block object at that specified index.

        :param item: The Block object or int to get the mapping data of
        :return: An int if a Block object was supplied, a Block object if an int was supplied
        """
        try:
            if isinstance(item, Block):
                return self._block_to_index_map[item]

            return self._index_to_block[item]
        except (KeyError, IndexError):
            raise KeyError(
                f"There is no {item} in the BlockManager. "
                f"You might want to use the `get_add_block` function for your blocks before accessing them."
            )

    def get_add_block(self, block: Block) -> int:
        """
        Adds a Block object to the internal Block object/ID mappings. If the Block already exists in the mappings,
        then the existing ID is returned

        :param block: The Block to add to the manager
        :return: The internal ID of the Block
        """
        assert isinstance(block, Block), "BlockManager only takes Block objects."
        if block in self._block_to_index_map:
            return self._block_to_index_map[block]

        self._block_to_index_map[block] = i = len(self._block_to_index_map)
        self._index_to_block.append(block)

        return i

    def register(self, block: Block) -> int:
        return self.get_add_block(block)



from collections import UserDict
from typing import Tuple, Iterable, Generator
import copy
import numpy

from amulet.api.block_entity import BlockEntity

Coordinate = Tuple[int, int, int]


class BlockEntityDict(UserDict):
    InputType = Iterable[BlockEntity]

    def __init__(self, block_entities: InputType = ()):
        super(BlockEntityDict, self).__init__()
        for block_entity in block_entities:
            self._assert_val(block_entity)
            self.data[block_entity.location] = block_entity

    def _assert_key(self, key):
        assert self._check_key(
            key
        ), f"Key must be in the format Tuple[int, int, int]. Got: {key}"

    @staticmethod
    def _check_key(key):
        return isinstance(key, tuple) and all(
            isinstance(a, (int, numpy.integer)) for a in key
        )

    def _assert_val(self, value):
        assert self._check_val(value), f"Val must be a BlockEntity. Got: {value}"

    @staticmethod
    def _check_val(value):
        return isinstance(value, BlockEntity)

    def __repr__(self) -> str:
        """ Return repr(self). """
        super_repr = (
            "".join(f"\n\t{key}:{val}" for key, val in self.data.items()) + "\n"
        )
        return f"BlockEntityDict({super_repr})"

    def clear(self) -> None:
        """ Remove all items from list. """
        self.data.clear()

    def keys(self) -> Generator[Coordinate, None, None]:
        yield from self.data.keys()

    def __iter__(self) -> Generator[BlockEntity, None, None]:
        yield from self.data.values()

    def values(self) -> Generator[BlockEntity, None, None]:
        yield from self.data.values()

    def items(self) -> Generator[Tuple[Coordinate, BlockEntity], None, None]:
        yield from self.data.items()

    def copy(self) -> "BlockEntityDict":
        return copy.deepcopy(self)

    def insert(self, block_entity: BlockEntity) -> None:
        """ Insert block_entity at its coordinates. """
        self._assert_val(block_entity)
        self.data[block_entity.location] = block_entity

    def pop(self, coordinate: Coordinate) -> BlockEntity:
        """
        Remove and return BlockEntity at coordinate.

        Raises IndexError if no BlockEntity at coordinate.
        """
        self._assert_key(coordinate)
        if coordinate in self.data:
            return self.data.pop(coordinate)
        raise IndexError

    def __delitem__(self, coordinate: Coordinate) -> None:
        """ Delete self[key]. """
        self._assert_key(coordinate)
        super().__delitem__(coordinate)

    def _check_block_entity(
        self, coordinate: Coordinate, block_entity: BlockEntity
    ) -> BlockEntity:
        self._assert_key(coordinate)
        self._assert_val(block_entity)
        if coordinate != block_entity.location:
            block_entity = block_entity.new_at_location(*coordinate)
        return block_entity

    def __getitem__(self, item: Coordinate) -> BlockEntity:
        return super().__getitem__(item)

    def __setitem__(self, coordinate: Coordinate, block_entity: BlockEntity) -> None:
        """ Set self[key] to value. """
        self.data[coordinate] = self._check_block_entity(coordinate, block_entity)

    def setdefault(
        self, coordinate: Coordinate, block_entity: BlockEntity
    ) -> BlockEntity:
        block_entity = self._check_block_entity(coordinate, block_entity)
        return self.data.setdefault(coordinate, block_entity)

    def popitem(self):
        raise NotImplementedError

    def update(self, block_entities: InputType) -> None:
        for block_entity in block_entities:
            self._assert_val(block_entity)
            self.data[block_entity.location] = block_entity



from __future__ import annotations

from .anvil_1912 import (
    Anvil1912Interface,
)


class Anvil1934Interface(Anvil1912Interface):
    def __init__(self):
        Anvil1912Interface.__init__(self)
        self.features["light_optional"] = "true"

    @staticmethod
    def minor_is_valid(key: int):
        return 1934 <= key < 2203


export = Anvil1934Interface



from __future__ import annotations

from .leveldb_7 import (
    LevelDB7Interface,
)


class LevelDB8Interface(LevelDB7Interface):
    def __init__(self):
        super().__init__()

        self.features["chunk_version"] = 8
        self.features["terrain"] = "2fnpalette"


export = LevelDB8Interface



from __future__ import annotations

from .leveldb_1 import (
    LevelDB1Interface,
)


class LevelDB2Interface(LevelDB1Interface):
    def __init__(self):
        super().__init__()

        self.features["chunk_version"] = 2


export = LevelDB2Interface
